from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class AmazonAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Amazon App"
    access_key_id = Attribute('Access Key ID', min_version='21.4')
    aws_credential_dn = Attribute('Aws Credential DN', min_version='21.4')
    binding_target = Attribute('Binding Target', min_version='21.4')
    certificate_arn = Attribute('Certificate ARN', min_version='21.4')
    certificate_name = Attribute('Certificate Name', min_version='21.4')
    cloudfront_distribution_id = Attribute('CloudFront Distribution ID', min_version='21.4')
    create_binding = Attribute('Create Binding', min_version='21.4')
    file_validation_disabled = Attribute('File Validation Disabled', min_version='21.4')
    iam_certificate_id = Attribute('IAM Certificate ID', min_version='21.4')
    initial_binding_attempt = Attribute('Initial Binding Attempt', min_version='21.4')
    install_path = Attribute('Install Path', min_version='21.4')
    issued_by_aws = Attribute('Issued By AWS', min_version='21.4')
    load_balancer_name = Attribute('Load Balancer Name', min_version='21.4')
    load_balancer_port = Attribute('Load Balancer Port', min_version='21.4')
    load_balancer_region_code = Attribute('Load Balancer Region Code', min_version='21.4')
    minimum_protocol_version = Attribute('Minimum Protocol Version', min_version='21.4')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    provisioning_to = Attribute('Provisioning To', min_version='21.4')
    region_code = Attribute('Region Code', min_version='21.4')
    replace_store = Attribute('Replace Store', min_version='21.4')
    secret_access_key = Attribute('Secret Access Key', min_version='21.4')
    target_group = Attribute('Target Group', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes

class ReportAnalyticsAttributes(ReportBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:Analytics"
    options = Attribute('Options', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes

class ThawteSPKICAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Thawte SPKI CA"
    certificate_block = Attribute('Certificate Block')
    certificate_transparency = Attribute('Certificate Transparency', min_version='16.3')
    retrieval_period = Attribute('Retrieval Period', min_version='15.4')
    server_type = Attribute('Server Type')
    signature_algorithm = Attribute('Signature Algorithm')
    uri = Attribute('URI')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class SANS(ObjectModel):
    name: str = ApiField(alias='Name')
    type_name: str = ApiField(alias='TypeName')

class PKI(ObjectModel):
    certificate_dn: str = ApiField(alias='CertificateDn')
    certificate_guid: str = ApiField(alias='CertificateGuid')
    pki_dn: str = ApiField(alias='PkiDn')
    pki_guid: str = ApiField(alias='PkiGuid')

class Certificate(ObjectModel):
    city: str = ApiField(alias='City')
    common_name: str = ApiField(alias='CommonName')
    country: str = ApiField(alias='Country')
    key_algorithm: str = ApiField(alias='KeyAlgorithm')
    key_bit_size: str = ApiField(alias='KeyBitSize')
    organization: str = ApiField(alias='Organization')
    organizational_units: list[str] = ApiField(alias='OrganizationalUnits', default_factory=list)
    sans: list[SANS] = ApiField(alias='Sans', default_factory=list)
    state: str = ApiField(alias='State')

class Installation(ObjectModel):
    credential_dn: str = ApiField(alias='CredentialDn')
    host: str = ApiField(alias='Host')
    port: str = ApiField(alias='Port')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.client_portal_base import ClientPortalBaseAttributes
from pyvenafi.tpp.attributes.client_work_base import ClientWorkBaseAttributes
from pyvenafi.tpp.attributes.x509_certificate_base import X509CertificateBaseAttributes

class ClientUserCertificateWorkAttributes(
    ClientPortalBaseAttributes,
    ClientWorkBaseAttributes,
    X509CertificateBaseAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Client User Certificate Work"
    certificate_bundle_capacity = Attribute('Certificate Bundle Capacity')
    certificate_bundle_capacity_mobile_config = Attribute('Certificate Bundle Capacity Mobile Config')
    certificate_container = Attribute('Certificate Container')
    certificate_icon = Attribute('Certificate Icon')
    download_instructions = Attribute('Download Instructions')
    download_limit = Attribute('Download Limit')
    include_historic_certificates = Attribute('Include Historic Certificates')
    membership_loss_disable = Attribute('Membership Loss Disable')
    membership_loss_revoke = Attribute('Membership Loss Revoke')
    naming_pattern = Attribute('Naming Pattern')
    outlook_profile_generation = Attribute('Outlook Profile Generation')
    outlook_profile_name = Attribute('Outlook Profile Name')
    outlook_profile_options = Attribute('Outlook Profile Options')
    portal_friendly_name = Attribute('Portal Friendly Name')
    publish_to_identity = Attribute('Publish To Identity')
    publish_to_identity_on_pre_enroll = Attribute('Publish To Identity on Pre-Enroll')
    required_member_identity = Attribute('Required Member Identity')
    transfer_allowed = Attribute('Transfer Allowed')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes
from pyvenafi.tpp.attributes.proxy import ProxyAttributes

class SymantecLHKCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
    __config_class__ = "Symantec LHK CA"
    fields = Attribute('Fields')
    uri = Attribute('URI')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes

class ReportEnhancedCertificateExpirationAttributes(ReportBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:EnhancedCertificateExpiration"
    debug_file = Attribute('Debug File', min_version='21.4')
    grouping = Attribute('Grouping', min_version='21.4')
    options = Attribute('Options', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.legacy_key_base import LegacyKeyBaseAttributes
from pyvenafi.tpp.attributes.monitoring_base import MonitoringBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class SymmetricKeyAttributes(LegacyKeyBaseAttributes, MonitoringBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Symmetric Key"
    check_value = Attribute('Check Value', min_version='21.4')



from __future__ import annotations

import json
from datetime import datetime
from typing import TYPE_CHECKING

import requests
import urllib3
from pydantic import BaseModel

if TYPE_CHECKING:
    from packaging.version import Version

class Session:
    """
    This class is responsible for holding the appropriate headers to authenticate each
    request. It also removes all null values from all data sent to TPP.
    """

    def __init__(
        self, headers: dict, proxies: dict = None, certificate_path: str = None,
        key_file_path: str = None, verify_ssl: bool = False, tpp_version: 'Version' = None,
        connection_timeout: float = None, read_timeout: float = None
    ):
        self.session = requests.Session()
        if not verify_ssl:
            # noinspection PyUnresolvedReferences
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.request_kwargs = {
            'headers': headers,
            'verify' : verify_ssl,
            'timeout': (connection_timeout, read_timeout)
        }
        if proxies:
            self.request_kwargs['proxies'] = proxies
        if bool(certificate_path) != bool(key_file_path):
            raise ValueError(
                'The path to both the certificate file and the key file must be '
                'given for certificate authentication.'
            )
        if certificate_path and key_file_path:
            self.request_kwargs['cert'] = (certificate_path, key_file_path)
        self.tpp_version = tpp_version

    def update_headers(self, new_headers):
        self.request_kwargs['headers'].update(new_headers)

    def delete(self, url: str, params: dict = None):
        """
        Sends a DELETE request to the given URL.

        Args:
            url: URL endpoint
            params: Dictionary of parameters to append to the URL

        Returns:
            Returns the raw response returned by the server.
        """
        return self.session.delete(url, params=self._sanitize(obj=params), **self.request_kwargs)

    def get(self, url: str, params: dict = None):
        """
        Sends a GET request to the given URL with the given URL parameters.

        Args:
            url: URL endpoint
            params: Dictionary of parameters to append to the URL

        Returns:
            Returns the raw response returned by the server.
        """
        return self.session.get(url=url, params=self._sanitize(obj=params), **self.request_kwargs)

    def post(self, url: str, data: dict):
        """
        Sends a POST request to the given URL with the given data.

        Args:
            url: URL endpoint
            data: Dictionary of data

        Returns:
            Returns the raw response returned by the server.
        """
        return self.session.post(url=url, data=self._to_json(obj=data), **self.request_kwargs)

    def put(self, url: str, data: dict):
        """
        Sends a PUT request to the given URL with the given data.

        Args:
            url: URL endpoint
            data: Dictionary of data

        Returns:
            Returns the raw response returned by the server.
        """
        return self.session.put(url=url, data=self._to_json(obj=data), **self.request_kwargs)

    def _to_json(self, obj):
        sanitized_obj = self._sanitize(obj)
        return json.dumps(sanitized_obj, default=self._json_default)

    @staticmethod
    def _json_default(obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return str(obj)

    @staticmethod
    def _validate_value(obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return obj

    def _sanitize(self, obj):
        if isinstance(obj, BaseModel):
            obj = obj.dict(by_alias=True, exclude_none=True)
        if isinstance(obj, dict):
            new_values = {}
            for k, v in obj.items():
                if v is None:
                    continue
                new_k = self._validate_value(k)
                new_v = self._sanitize(v)
                new_values[new_k] = new_v
            return new_values
        elif isinstance(obj, (list, tuple, set)):
            return [self._sanitize(item) for item in obj if item is not None]
        else:
            return self._validate_value(obj)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.log_channel import LogChannelAttributes

class LogSNMPAttributes(LogChannelAttributes, metaclass=IterableMeta):
    __config_class__ = "Log SNMP"
    community = Attribute('Community', min_version='21.4')
    message_body = Attribute('Message Body', min_version='21.4')
    snmp_authentication_password = Attribute('SNMP Authentication Password', min_version='21.4')
    snmp_authentication_protocol = Attribute('SNMP Authentication Protocol', min_version='21.4')
    snmp_engine_id = Attribute('SNMP Engine ID', min_version='21.4')
    snmp_privacy_password = Attribute('SNMP Privacy Password', min_version='21.4')
    snmp_privacy_protocol = Attribute('SNMP Privacy Protocol', min_version='21.4')
    snmp_username = Attribute('SNMP Username', min_version='21.4')
    snmp_v3_mode = Attribute('SNMP V3 Mode', min_version='21.4')
    snmp_version = Attribute('SNMP Version', min_version='21.4')
    target = Attribute('Target', min_version='21.4')
    trap_oid = Attribute('Trap OID', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes
from pyvenafi.tpp.attributes.report_filter_base import ReportFilterBaseAttributes

class ReportCustomReportAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:Custom Report"
    custom_report_data_type = Attribute('Custom Report: Data Type', min_version='21.4')
    custom_report_entity_type = Attribute('Custom Report: Entity Type', min_version='21.4')
    custom_report_filter = Attribute('Custom Report: Filter', min_version='21.4')
    custom_report_order_by = Attribute('Custom Report: Order By', min_version='21.4')
    custom_report_owner = Attribute('Custom Report: Owner', min_version='21.4')
    custom_report_personalized = Attribute('Custom Report: Personalized', min_version='21.4')
    custom_report_selected_field = Attribute('Custom Report: Selected Field', min_version='21.4')
    custom_report_sort_descending = Attribute('Custom Report: Sort Descending', min_version='21.4')
    error_message = Attribute('Error Message', min_version='21.4')
    options = Attribute('Options', min_version='21.4')
    status = Attribute('Status', min_version='21.4')
    summary = Attribute('Summary', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.key_pair import KeyPairAttributes

class SSHCAKeyPairAttributes(KeyPairAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH CA Key Pair"
    fingerprint_sha256 = Attribute('Fingerprint SHA256')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.log_channel import LogChannelAttributes

class LogFileAttributes(LogChannelAttributes, metaclass=IterableMeta):
    __config_class__ = "Log File"
    expiration = Attribute('Expiration', min_version='21.4')
    language = Attribute('Language', min_version='21.4')
    log_directory = Attribute('Log Directory', min_version='21.4')
    max_fileage = Attribute('Max Fileage', min_version='21.4')
    max_filesize = Attribute('Max Filesize', min_version='21.4')
    translate = Attribute('Translate', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class EncryptionRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Encryption Root"
    compression_level = Attribute('Compression Level', min_version='22.3')
    encryption_driver = Attribute('Encryption Driver', min_version='21.4')
    protection_key = Attribute('Protection Key', min_version='21.4')
    rotation_log = Attribute('Rotation Log', min_version='24.1')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    CloudApiEndpoint,
    CloudApiOutputModel,
    generate_output,
)
from pyvenafi.cloud.api.models import tagging_service

class _tagging_service:
    def __init__(self, api_obj):
        self.v1 = self._v1(api_obj=api_obj)

    class _v1(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='v1')
            self.tags = self._tags(api_obj=self._api_obj, url=f'{self._url}/tags')
            self.tagsassignment = self._tagsassignment(api_obj=self._api_obj, url=f'{self._url}/tagsassignment')

        class _tags(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.creation = self._creation(api_obj=self._api_obj, url=f'{self._url}/creation')
                self.deletion = self._deletion(api_obj=self._api_obj, url=f'{self._url}/deletion')
                self.values = self._values(api_obj=self._api_obj, url=f'{self._url}/values')

            def NAME(self, name: str):
                return self._NAME(api_obj=self._api_obj, url=f'{self._url}/{name}')

            def get(self):
                class Output(CloudApiOutputModel):
                    TagResponse: tagging_service.TagResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'TagResponse'
                    }
                )

            def post(self, TagRequest: tagging_service.TagRequest):
                data = {**TagRequest.dict()}

                class Output(CloudApiOutputModel):
                    TagInformation: tagging_service.TagInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        201: 'TagInformation'
                    }
                )

            class _NAME(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.values = self._values(api_obj=self._api_obj, url=f'{self._url}/values')

                def delete(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._delete(params={}))

                def get(self):
                    class Output(CloudApiOutputModel):
                        TagInformation: tagging_service.TagInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'TagInformation'
                        }
                    )

                class _values(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def VALUE(self, value: str):
                        return self._VALUE(api_obj=self._api_obj, url=f'{self._url}/{value}')

                    def get(self):
                        class Output(CloudApiOutputModel):
                            TagValuesResponse: tagging_service.TagValuesResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'TagValuesResponse'
                            }
                        )

                    def post(self, TagValuesRequest: tagging_service.TagValuesRequest):
                        data = {**TagValuesRequest.dict()}

                        class Output(CloudApiOutputModel):
                            TagValuesResponse: tagging_service.TagValuesResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                201: 'TagValuesResponse'
                            }
                        )

                    class _VALUE(CloudApiEndpoint):
                        def delete(self):
                            class Output(CloudApiOutputModel):
                                pass

                            return generate_output(output_cls=Output, response=self._delete(params={}))

            class _creation(CloudApiEndpoint):
                def post(self, TagsBulkRequest: tagging_service.TagsBulkRequest):
                    data = {**TagsBulkRequest.dict()}

                    class Output(CloudApiOutputModel):
                        TagsAndValuesResponse: tagging_service.TagsAndValuesResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'TagsAndValuesResponse'
                        }
                    )

            class _deletion(CloudApiEndpoint):
                def post(self, TagsBulkRequest: tagging_service.TagsBulkRequest):
                    data = {**TagsBulkRequest.dict()}

                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._post(data=data))

            class _values(CloudApiEndpoint):
                def get(self):
                    class Output(CloudApiOutputModel):
                        TagValuesResponse: tagging_service.TagValuesResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'TagValuesResponse'
                        }
                    )

        class _tagsassignment(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.aggregates = self._aggregates(api_obj=self._api_obj, url=f'{self._url}/aggregates')

            def patch(self, TagsAssignRequest: tagging_service.TagsAssignRequest):
                data = {**TagsAssignRequest.dict()}

                class Output(CloudApiOutputModel):
                    TagsAssignResponse: tagging_service.TagsAssignResponse

                return generate_output(
                    output_cls=Output,
                    response=self._patch(data=data),
                    rc_mapping={
                        200: 'TagsAssignResponse'
                    }
                )

            class _aggregates(CloudApiEndpoint):
                def post(self, TagsAssignmentAggregatesRequest: tagging_service.TagsAssignmentAggregatesRequest):
                    data = {**TagsAssignmentAggregatesRequest.dict()}

                    class Output(CloudApiOutputModel):
                        TagsAssignmentAggregatesResponse: tagging_service.TagsAssignmentAggregatesResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            200: 'TagsAssignmentAggregatesResponse'
                        }
                    )



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ReportTemplateAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Report Template"
    item_vault_id = Attribute('Item Vault Id')
    report_class = Attribute('Report Class')
    template_vault_id = Attribute('Template Vault Id')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class UserPreferenceAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "User Preference"
    console_timeout = Attribute('Console Timeout', min_version='21.4')
    user_preferences = Attribute('User Preferences', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ApplicationGroupAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Application Group"
    certificate = Attribute('Certificate')
    common_data_location = Attribute('Common Data Location')
    common_data_vault_id = Attribute('Common Data Vault Id')
    enrollment_application_dn = Attribute('Enrollment Application DN')
    primary_application_dn = Attribute('Primary Application DN')
    private_key_stub_vault_id = Attribute('Private Key Stub Vault Id')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class RemoteAccessRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Remote Access Root"
    approved_issuer = Attribute('Approved Issuer', min_version='21.4')
    attempt_sid_extension_before_source_field = Attribute(
        'Attempt SID Extension Before Source Field',
        min_version='23.3'
    )
    authentication_methods = Attribute('Authentication Methods', min_version='21.4')
    certificate_auth_lookup_field = Attribute('Certificate Auth Lookup Field', min_version='21.4')
    certificate_auth_source_field = Attribute('Certificate Auth Source Field', min_version='21.4')
    certificate_auth_source_regex = Attribute('Certificate Auth Source Regex', min_version='21.4')
    certificate_pinning_options = Attribute('Certificate Pinning Options', min_version='21.4')
    certificate_pinning_work_dn = Attribute('Certificate Pinning Work DN', min_version='21.4')
    expiration = Attribute('Expiration', min_version='21.4')
    grant_validity = Attribute('Grant Validity', min_version='21.4')
    jwt_auth_source_field = Attribute('Jwt Auth Source Field', min_version='22.4')
    mapping_table = Attribute('Mapping Table', min_version='21.4')
    oauth_device_authorization_verification_uri = Attribute(
        'OAuth Device Authorization Verification URI',
        min_version='22.1'
    )
    oauth_device_authorization_verification_uri_complete_format = Attribute(
        'OAuth Device Authorization Verification URI Complete Format',
        min_version='22.1'
    )
    openapi_documentation = Attribute('OpenApi Documentation', min_version='22.4')
    refresh_api_enabled = Attribute('Refresh API Enabled', min_version='21.4')
    refresh_token_enabled = Attribute('Refresh Token Enabled', min_version='21.4')
    session_expiration_interval = Attribute('Session Expiration Interval', min_version='22.4')
    session_pool_entries = Attribute('Session Pool Entries', min_version='21.4')
    session_pool_max_age = Attribute('Session Pool Max Age', min_version='21.4')
    session_rights_refresh_age = Attribute('Session Rights Refresh Age', min_version='21.4')
    token_validity = Attribute('Token Validity', min_version='21.4')
    unused_token_expiration = Attribute('Unused Token Expiration', min_version='22.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.encryption_driver import EncryptionDriverAttributes

class Pkcs11EncryptionDriverAttributes(EncryptionDriverAttributes, metaclass=IterableMeta):
    __config_class__ = "Pkcs11 Encryption Driver"
    account_type = Attribute('Account Type', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    cryptokipath = Attribute('CryptokiPath', min_version='21.4')
    dsn = Attribute('DSN', min_version='21.4')
    default_key = Attribute('Default Key', min_version='21.4')
    encrypted_pin = Attribute('Encrypted Pin', min_version='21.4')
    fallback_pin = Attribute('Fallback Pin', min_version='21.4')
    key_derivation = Attribute('Key Derivation', min_version='22.1')
    key_generation = Attribute('Key Generation', min_version='21.4')
    key_storage = Attribute('Key Storage', min_version='21.4')
    key_validation = Attribute('Key Validation', min_version='21.4')
    permitted_keys = Attribute('Permitted Keys', min_version='21.4')
    slot_id = Attribute('Slot Id', min_version='21.4')
    token_label = Attribute('Token Label', min_version='21.4')
    token_serial = Attribute('Token Serial', min_version='21.4')
    verigram = Attribute('VeriGram', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class ValidationBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Validation Base"
    file_validation_error = Attribute('File Validation Error', min_version='21.4')
    file_validation_result = Attribute('File Validation Result', min_version='21.4')
    last_validation = Attribute('Last Validation', min_version='21.4')
    last_validation_result = Attribute('Last Validation Result', min_version='21.4')
    ssl_listen_host = Attribute('SSL Listen Host', min_version='21.4')
    ssl_listen_port = Attribute('SSL Listen Port', min_version='21.4')
    use_specified_host = Attribute('Use Specified Host', min_version='21.4')
    validation_disabled = Attribute('Validation Disabled', min_version='21.4')
    validation_errors = Attribute('Validation Errors', min_version='21.4')
    validation_results = Attribute('Validation Results', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes

class CertificateAuthorityBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Certificate Authority Base"
    additional_field = Attribute('Additional Field')
    concurrent_connection_limit = Attribute('Concurrent Connection Limit')
    credential = Attribute('Credential')
    credits = Attribute('Credits')
    credits_alert = Attribute('Credits Alert')
    credits_used = Attribute('Credits Used')
    enhanced_key_usage = Attribute('Enhanced Key Usage')
    host = Attribute('Host')
    manual_approval = Attribute('Manual Approval')
    port = Attribute('Port')
    protection_key = Attribute('Protection Key')
    renewal_window = Attribute('Renewal Window')
    retrieval_period = Attribute('Retrieval Period')
    retry_count = Attribute('Retry Count')
    retry_interval = Attribute('Retry Interval')
    san_enabled = Attribute('SAN Enabled')
    signature_algorithm = Attribute('Signature Algorithm')
    specific_end_date_enabled = Attribute('Specific End Date Enabled')
    template = Attribute('Template')
    test_account = Attribute('Test Account')
    timeout = Attribute('Timeout')
    validity_period = Attribute('Validity Period')
    vault_id = Attribute('Vault Id')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import log

class _Log(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Log')
        self.LogSchema = self._LogSchema(api_obj=api_obj, url=f'{self._url}/LogSchema')
        self.EvaluateMacro = self._EvaluateMacro(api_obj=api_obj, url=f'{self._url}/EvaluateMacro')
        self.AdaptableInfo = self._AdaptableInfo(api_obj=api_obj, url=f'{self._url}/AdaptableInfo')

    def get(
        self, component: str = None, from_time: str = None, grouping: int = None, id: int = None,
        limit: int = None, offset: int = None, order: str = None, severity: str = None, text1: str = None,
        text2: str = None, to_time: str = None, value1: str = None, value2: str = None
    ):
        params = {
            'Component': component,
            'FromTime' : from_time,
            'Grouping' : grouping,
            'Id'       : id,
            'Limit'    : limit,
            'Offset'   : offset,
            'Order'    : order,
            'Severity' : severity,
            'Text1'    : text1,
            'Text2'    : text2,
            'ToTime'   : to_time,
            'Value1'   : value1,
            'Value2'   : value2
        }

        class Output(WebSdkOutputModel):
            log_events: list[log.LogEvent] = ApiField(default_factory=list, alias='LogEvents')

        return generate_output(output_cls=Output, response=self._get(params=params))

    def post(
        self, component: str, id: int, grouping: int = None, severity: int = None, source_ip: str = None,
        text1: str = None, text2: str = None, value1: str = None, value2: str = None
    ):
        body = {
            'Component': component,
            'ID'       : id,
            'Grouping' : grouping,
            'Severity' : severity,
            'SourceIp' : source_ip,
            'Text1'    : text1,
            'Text2'    : text2,
            'Value1'   : value1,
            'Value2'   : value2
        }

        class Output(WebSdkOutputModel):
            log_result: int = ApiField(alias='LogResult')

        return generate_output(output_cls=Output, response=self._post(data=body))

    def Guid(self, guid: str):
        return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

    class _Guid(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                log_events: list[log.LogEvent] = ApiField(default_factory=list, alias='LogEvents')

            return generate_output(output_cls=Output, response=self._get())

    class _LogSchema(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                log_event_application_definitions: list[log.LogEventApplicationDefinition] = ApiField(
                    alias='LogEventApplicationDefinitions'
                )
                log_event_definitions: list[log.LogEventDefinition] = ApiField(
                    alias='LogEventDefinitions',
                    default_factory=list
                )
                log_result: int = ApiField(alias='LogResult')

            return generate_output(output_cls=Output, response=self._get())

    class _AdaptableInfo(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                scripts: list[log.Script] = ApiField(alias='Scripts')
                script_errors: list[str] = ApiField(alias='ScriptErrors')

            return generate_output(output_cls=Output, response=self._get())

    class _EvaluateMacro(WebSdkEndpoint):
        def post(
            self,
            macro_source: str = None,
            id: int = None,
            severity: int = None,
            component: str = None,
            text_1: str = None,
            text_2: str = None,
            value_1: int = None,
            value_2: int = None,
        ):
            body = {
                "MacroSource": macro_source,
                "ID"         : id,
                "Severity"   : severity,
                "Component"  : component,
                "Text1"      : text_1,
                "Text2"      : text_2,
                "Value1"     : value_1,
                "Value2"     : value_2,
            }

            class Output(WebSdkOutputModel):
                macro_result: str = ApiField(alias='MacroResult')
                log_result: int = ApiField(alias='LogResult')

            return generate_output(output_cls=Output, response=self._post(data=body))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.flow_action_base import FlowActionBaseAttributes

class CodeSignPreApprovalActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Code Sign PreApproval Action"
    stage_identifier = Attribute('Stage Identifier', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class AdaptableAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Adaptable App"
    allow_push_without_private_key = Attribute('Allow Push Without Private Key', min_version='21.4')
    certificate_name = Attribute('Certificate Name', min_version='21.4')
    file_validation_disabled = Attribute('File Validation Disabled', min_version='21.4')
    installation_status = Attribute('Installation Status', min_version='21.4')
    log_debug = Attribute('Log Debug', min_version='21.4')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    option_1 = Attribute('Option 1', min_version='21.4')
    option_2 = Attribute('Option 2', min_version='21.4')
    pbes2_algorithm = Attribute('PBES2 Algorithm', min_version='21.4')
    pk_credential = Attribute('PK Credential', min_version='21.4')
    password_1 = Attribute('Password 1', min_version='21.4')
    powershell_script = Attribute('PowerShell Script', min_version='21.4')
    powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.4')
    retry_after_script_hash_mismatch = Attribute('Retry After Script Hash Mismatch', min_version='21.4')
    script_execution_timeout = Attribute('Script Execution Timeout', min_version='21.4')
    script_hash_mismatch_error = Attribute('Script Hash Mismatch Error', min_version='21.4')
    secondary_credential = Attribute('Secondary Credential', min_version='21.4')
    text_field_1 = Attribute('Text Field 1', min_version='21.4')
    text_field_2 = Attribute('Text Field 2', min_version='21.4')
    text_field_3 = Attribute('Text Field 3', min_version='21.4')
    text_field_4 = Attribute('Text Field 4', min_version='21.4')
    text_field_5 = Attribute('Text Field 5', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.code_signing_environment_base import CodeSigningEnvironmentBaseAttributes

class CodeSigningCSPEnvironmentAttributes(CodeSigningEnvironmentBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Code Signing CSP Environment"
    encryption_key_dn = Attribute('Encryption Key DN', min_version='21.4')
    signing_key_dn = Attribute('Signing Key DN', min_version='21.4')



from __future__ import annotations

from typing import Union

from pyvenafi.tpp.api.websdk.enums.config import CodeSignAttributeValues
from pyvenafi.tpp.api.websdk.models.codesign import (
    AppleEnvironment,
    AppleTemplate,
    CertificateEnvironment,
    CertificateTemplate,
    CSPEnvironment,
    CSPTemplate,
    DotNetEnvironment,
    DotNetTemplate,
    GPGEnvironment,
    GPGTemplate,
    KeyPairEnvironment,
    KeyPairTemplate,
    Project,
)
from pyvenafi.tpp.api.websdk.models.config import Object
from pyvenafi.tpp.features.bases.feature_base import (
    feature,
    FeatureBase,
)

TemplateTypes = Union[
    AppleTemplate,
    CertificateTemplate,
    CSPTemplate,
    DotNetTemplate,
    GPGTemplate,
    KeyPairTemplate,
]

EnvironmentTypes = Union[
    AppleEnvironment,
    CertificateEnvironment,
    CSPEnvironment,
    DotNetEnvironment,
    GPGEnvironment,
    KeyPairEnvironment,
]

@feature('Environment')
class CodeSignEnvironment(FeatureBase):
    def create(
        self,
        name: str,
        template: Union[str, Object, TemplateTypes],
        project: Union[str, Object, Project],
        raise_if_already_exists: bool = False,
    ) -> EnvironmentTypes:
        """
        Create an environment for a project using a template.

        Args:
            name: Name of the environment.
            template: :class:`~.models.codesign.Template`.
            project: :class:`~.models.codesign.Project`.
            raise_if_already_exists: If the object already exists, raise an error.

        Returns:
            A CodeSign Environment
        """

        project_dn = self._get_dn(project)
        project = self._api.websdk.Codesign.GetProject.post(dn=project_dn).project

        template_dn = self._get_dn(template)
        template_output = self._api.websdk.Codesign.GetTemplate.post(dn=template_dn)
        template_types = {
            CodeSignAttributeValues.TemplateTypeKey.apple_template      : template_output.apple_template,
            CodeSignAttributeValues.TemplateTypeKey.certificate_template: template_output.certificate_template,
            CodeSignAttributeValues.TemplateTypeKey.csp_template        : template_output.csp_template,
            CodeSignAttributeValues.TemplateTypeKey.dotnet_template     : template_output.dot_net_template,
            CodeSignAttributeValues.TemplateTypeKey.gpg_template        : template_output.gpg_template,
            CodeSignAttributeValues.TemplateTypeKey.key_pair_template   : template_output.key_pair_template
        }
        try:
            template_type, template = next((k, v) for k, v in template_types.items() if v)
        except:
            raise ValueError('The template provided does not exist.')
        if self._is_version_compatible(minimum='24.3'):
            dn = rf'{project_dn}\{name}'
        else:
            template_dn = template_dn.replace("\\", "\\\\")
            project_dn = project_dn.replace("\\", "\\\\")
            dn = rf'{project_dn}\\{name}'

        try:
            create_output = self._api.websdk.Codesign.CreateEnvironment.post(
                dn=dn,
                project={
                    'Dn'  : project_dn,
                    'Guid': self._get_guid(project),
                    'Id'  : project.id
                },
                environment_name=name,
                template_dn=template_dn,
                template_type=template_type,
                template=template
            )
            create_output.assert_valid_response()
        except Exception as e:
            if raise_if_already_exists:
                raise e

        environment = self.get(dn=dn)
        return environment

    def update(
        self,
        environment: Union[Object, EnvironmentTypes]
    ):
        """
        Update an environment.

        Args:
            environment: The updated environment.
        """
        if isinstance(environment, AppleEnvironment):
            key = 'apple_environment'
        elif isinstance(environment, CertificateEnvironment):
            key = 'certificate_environment'
        elif isinstance(environment, CSPEnvironment):
            key = 'csp_environment'
        elif isinstance(environment, DotNetEnvironment):
            key = 'dot_net_environment'
        elif isinstance(environment, GPGEnvironment):
            key = 'gpg_environment'
        elif isinstance(environment, KeyPairEnvironment):
            key = 'key_pair_environment'
        else:
            raise ValueError(f'Invalid type {type(environment)}')
        kwargs = {
            key: environment
        }
        output = self._api.websdk.Codesign.UpdateEnvironment.post(**kwargs)
        output.assert_valid_response()

    def get(
        self,
        dn: str = None,
        guid: str = None,
        id: str = None
    ) -> EnvironmentTypes:
        """
        Get an environment using a :ref:`dn`, :ref:`guid`, or Identifier.

        Args:
            dn: :ref:`dn` of the environment.
            guid: :ref:`guid` of the environment.
            id: Identifier of the environment.

        Returns:
            A CodeSign environment.
        """
        if any([dn, guid, id]):
            output = self._api.websdk.Codesign.GetEnvironment.post(dn=dn, guid=guid, id=id)
            output.assert_valid_response()
        else:
            raise ValueError('You must provide a dn, guid, or id to get an environment.')
        environment = (output.apple_environment or output.certificate_environment or output.csp_environment or
                       output.dot_net_environment or output.gpg_environment or output.key_pair_environment)
        if not environment:
            raise Exception('Environment does not exist.')
        return environment

    def delete(
        self,
        environment: Union[str, Object, EnvironmentTypes]
    ):
        """
        Delete an environment.

        Args:
            environment: Environment, :ref:`dn` , or :ref:`config_object` of the environment.
        """
        dn = self._get_dn(environment)
        if dn:
            self._api.websdk.Codesign.DeleteEnvironment.post(dn=dn)
        else:
            guid = self._get_guid(environment)
            if guid:
                self._api.websdk.Codesign.DeleteEnvironment.post(guid=guid)
            else:
                raise ValueError('Invalid argument to delete the environment.')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes

class GoogleCloudCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Google Cloud CA"
    certificate_authority_id = Attribute('Certificate Authority ID', min_version='21.4')
    google_project_name = Attribute('Google Project Name', min_version='22.3')
    region = Attribute('Region', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes

class MicrosoftCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Microsoft CA"
    enrollment_agent_certificate = Attribute('Enrollment Agent Certificate', min_version='21.4')
    given_name = Attribute('Given Name', min_version='21.4')
    include_cn_as_san = Attribute('Include CN as SAN', min_version='21.4')
    use_external_msca_communicator = Attribute('Use External MSCA Communicator', min_version='22.1')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class WorkflowTicketAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Workflow Ticket"
    approval_explanation = Attribute('Approval Explanation', min_version='21.4')
    approval_from = Attribute('Approval From', min_version='21.4')
    approval_reason = Attribute('Approval Reason', min_version='21.4')
    approver_not_found_timestamp = Attribute('Approver Not Found Timestamp', min_version='21.4')
    creation_date = Attribute('Creation Date', min_version='21.4')
    last_notification = Attribute('Last Notification', min_version='21.4')
    last_update = Attribute('Last Update', min_version='21.4')
    notified_on = Attribute('Notified On', min_version='24.1')
    owner_object = Attribute('Owner Object', min_version='21.4')
    scheduled_start = Attribute('Scheduled Start', min_version='21.4')
    scheduled_stop = Attribute('Scheduled Stop', min_version='21.4')
    status = Attribute('Status', min_version='21.4')
    suspended_attribute = Attribute('Suspended Attribute', min_version='21.4')
    updated_by = Attribute('Updated By', min_version='21.4')
    user_data = Attribute('User Data', min_version='21.4')
    workflow = Attribute('Workflow', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class iPlanetAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "iPlanet"
    alias = Attribute('Alias', min_version='21.4')
    certutil_path = Attribute('Certutil Path', min_version='21.4')
    create_store = Attribute('Create Store', min_version='21.4')
    create_virtual_server = Attribute('Create Virtual Server', min_version='21.4')
    database_credential = Attribute('Database Credential', min_version='21.4')
    database_prefix = Attribute('Database Prefix', min_version='21.4')
    database_type = Attribute('Database Type', min_version='21.4')
    database_validation_disabled = Attribute('Database Validation Disabled', min_version='21.4')
    document_root = Attribute('Document Root', min_version='21.4')
    file_validation_disabled = Attribute('File Validation Disabled', min_version='21.4')
    install_path = Attribute('Install Path', min_version='21.4')
    key_store = Attribute('Key Store', min_version='21.4')
    key_store_credential = Attribute('Key Store Credential', min_version='21.4')
    mta_host = Attribute('MTA Host', min_version='21.4')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    pk12util_path = Attribute('Pk12util Path', min_version='21.4')
    protocol = Attribute('Protocol', min_version='21.4')
    replace_store = Attribute('Replace Store', min_version='21.4')
    secure_server_name = Attribute('Secure Server Name', min_version='21.4')
    use_proxy = Attribute('Use Proxy', min_version='21.4')
    virtual_server_dns_value = Attribute('Virtual Server DNS Value', min_version='21.4')
    virtual_server_port = Attribute('Virtual Server Port', min_version='21.4')
    virtual_server_user = Attribute('Virtual Server User', min_version='21.4')
    web_credential = Attribute('Web Credential', min_version='21.4')
    web_port = Attribute('Web Port', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.device import DeviceAttributes

class SSHManualInstallKeyDeviceAttributes(DeviceAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH Manual Install Key Device"
    ssh_manual_install_key_contact_email = Attribute('SSH Manual Install Key Contact Email')
    ssh_manual_install_key_location = Attribute('SSH Manual Install Key Location')
    ssh_manual_install_key_notes = Attribute('SSH Manual Install Key Notes')
    ssh_manual_install_key_owner = Attribute('SSH Manual Install Key Owner')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.client_certificate_work import ClientCertificateWorkAttributes

class NetworkDeviceCertificateWorkAttributes(ClientCertificateWorkAttributes, metaclass=IterableMeta):
    __config_class__ = "Network Device Certificate Work"
    authentication_credentials = Attribute('Authentication Credentials', min_version='21.4')
    ca_template_trust_anchors_enabled = Attribute('CA Template Trust Anchors Enabled', min_version='21.4')
    certificates_distribution_type = Attribute('Certificates Distribution Type', min_version='21.4')
    client_certificate_eku_checks_enabled = Attribute('Client Certificate Eku Checks Enabled', min_version='21.4')
    explicit_trust_anchors = Attribute('Explicit Trust Anchors', min_version='21.4')
    fallback_to_http_auth = Attribute('Fallback To Http Auth', min_version='21.4')
    http_basic_auth_disabled = Attribute('Http Basic Auth Disabled', min_version='21.4')
    http_digest_auth_disabled = Attribute('Http Digest Auth Disabled', min_version='21.4')
    pop_mode = Attribute('PoP Mode', min_version='21.4')
    reenrollment_subset_subject_matching_disabled = Attribute(
        'ReEnrollment Subset Subject Matching Disabled',
        min_version='21.4'
    )
    require_additional_http_auth = Attribute('Require Additional Http Auth', min_version='21.4')
    retryafter_disabled = Attribute('RetryAfter Disabled', min_version='21.4')
    revocation_check_timeout = Attribute('Revocation Check Timeout', min_version='21.4')
    revocation_mode = Attribute('Revocation Mode', min_version='21.4')
    revoke_existing_certificate_on_reenrollment = Attribute(
        'Revoke Existing Certificate On ReEnrollment',
        min_version='21.4'
    )
    revoke_existing_certificate_on_reenrollment_delay = Attribute(
        'Revoke Existing Certificate On ReEnrollment Delay',
        min_version='21.4'
    )
    use_implicit_trust_anchors = Attribute('Use Implicit Trust Anchors', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes
from pyvenafi.tpp.attributes.proxy import ProxyAttributes

class AdaptableCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
    __config_class__ = "Adaptable CA"
    allow_reissue = Attribute('Allow Reissue', min_version='21.4')
    certificate_credential = Attribute('Certificate Credential', min_version='21.4')
    connection_valid = Attribute('Connection Valid', min_version='21.4')
    custom_fields = Attribute('Custom Fields', min_version='21.4')
    interoperability_script = Attribute('Interoperability Script', min_version='21.4')
    log_debug = Attribute('Log Debug', min_version='21.4')
    oauth_token_application_id = Attribute('OAuth Token Application Id', min_version='22.1')
    oauth_token_credential = Attribute('OAuth Token Credential', min_version='22.1')
    oauth_token_scope = Attribute('OAuth Token Scope', min_version='22.1')
    powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.4')
    profile_string = Attribute('Profile String', min_version='21.4')
    renewal_window = Attribute('Renewal Window', min_version='21.4')
    retry_after_script_hash_mismatch = Attribute('Retry After Script Hash Mismatch', min_version='21.4')
    script_execution_timeout = Attribute('Script Execution Timeout', min_version='21.4')
    secondary_credential = Attribute('Secondary Credential', min_version='21.4')
    service_address = Attribute('Service Address', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)

class _ServerStatus(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/ServerStatus')
        self.Idle = self._Idle(api_obj=self._api_obj, url=f'{self._url}/Idle')
        self.Active = self._Active(api_obj=self._api_obj, url=f'{self._url}/Active')

    class _Idle(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                pass

            return generate_output(output_cls=Output, response=self._get())

    class _Active(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                pass

            return generate_output(output_cls=Output, response=self._get())



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.connection_base import ConnectionBaseAttributes
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes
from pyvenafi.tpp.attributes.validation_base import ValidationBaseAttributes

class ApplicationBaseAttributes(
    ConnectionBaseAttributes,
    DriverBaseAttributes,
    ValidationBaseAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Application Base"
    adaptable_workflow_approvers = Attribute('Adaptable Workflow Approvers')
    adaptable_workflow_reference_id = Attribute('Adaptable Workflow Reference ID')
    adaptable_workflow_stage = Attribute('Adaptable Workflow Stage')
    agent_validate_now = Attribute('Agent Validate Now')
    approver = Attribute('Approver')
    certificate = Attribute('Certificate')
    certificate_file = Attribute('Certificate File')
    certificate_installed = Attribute('Certificate Installed')
    created_by = Attribute('Created By')
    discovered_by_dn = Attribute('Discovered By DN')
    discovered_on = Attribute('Discovered On')
    file_owner_group = Attribute('File Owner: Group')
    file_owner_user = Attribute('File Owner: User')
    file_permissions_enabled = Attribute('File Permissions Enabled')
    file_permissions_group = Attribute('File Permissions: Group')
    file_permissions_user = Attribute('File Permissions: User')
    grouping_id = Attribute('Grouping Id')
    in_error = Attribute('In Error')
    in_process = Attribute('In Process')
    key_encryption_algorithm = Attribute('Key Encryption Algorithm')
    key_store_vault_id = Attribute('Key Store Vault Id')
    last_pushed_by = Attribute('Last Pushed By')
    last_pushed_on = Attribute('Last Pushed On')
    placement_id = Attribute('Placement Id')
    private_key_password_credential = Attribute('Private Key Password Credential')
    remote_one_to_many_generation = Attribute('Remote One To Many Generation')
    restart_application = Attribute('Restart Application')
    stage = Attribute('Stage')
    status = Attribute('Status')
    ticket_dn = Attribute('Ticket DN')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class AgentSSHBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Agent SSH Base"
    credential = Attribute('Credential', min_version='21.4')
    exclude_remote_mount_points = Attribute('Exclude Remote Mount Points', min_version='21.4')
    expiration = Attribute('Expiration', min_version='21.4')
    file_scan_disabled = Attribute('File Scan Disabled', min_version='21.4')
    log_threshold = Attribute('Log Threshold', min_version='21.4')
    max_filesize = Attribute('Max Filesize', min_version='21.4')
    min_filesize = Attribute('Min Filesize', min_version='21.4')
    port_validation_disabled = Attribute('Port Validation Disabled', min_version='21.4')
    ssh_scanner_service_path = Attribute('SSH Scanner Service Path', min_version='21.4')
    ssh_scanner_user_path = Attribute('SSH Scanner User Path', min_version='21.4')
    sshd_filename_filter = Attribute('SSHd Filename Filter', min_version='21.4')
    sshd_max_filesize = Attribute('SSHd Max Filesize', min_version='21.4')
    server_path_defaults_disabled = Attribute('Server Path Defaults Disabled', min_version='21.4')
    upload_private_keys_disabled = Attribute('Upload Private Keys Disabled', min_version='21.4')
    user_path_defaults_disabled = Attribute('User Path Defaults Disabled', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class JuniperSASAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Juniper SAS"
    external_port = Attribute('External Port')
    file_validation_disabled = Attribute('File Validation Disabled')
    internal_port = Attribute('Internal Port')
    network_validation_disabled = Attribute('Network Validation Disabled')
    reassign_ports = Attribute('Reassign Ports')
    vlan_port = Attribute('Vlan Port')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ACMEOrderAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "ACME Order"
    acme_authorization_dn = Attribute('ACME Authorization DN', min_version='21.4')
    acme_expires = Attribute('ACME Expires', min_version='21.4')
    acme_not_after = Attribute('ACME Not After', min_version='21.4')
    acme_not_before = Attribute('ACME Not Before', min_version='21.4')
    csr_vault_id = Attribute('CSR Vault Id', min_version='21.4')
    certificate = Attribute('Certificate', min_version='21.4')
    error = Attribute('Error', min_version='21.4')
    identifier = Attribute('Identifier', min_version='21.4')
    status = Attribute('Status', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ClientGroupBaseAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Client Group Base"
    assigned_work = Attribute('Assigned Work', min_version='21.4')
    client_portal_access_identity = Attribute('Client Portal Access Identity', min_version='21.4')
    fixed_members = Attribute('Fixed Members', min_version='21.4')
    in_error = Attribute('In Error', min_version='21.4')
    rank = Attribute('Rank', min_version='21.4')
    rule = Attribute('Rule', min_version='21.4')
    status = Attribute('Status', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.client_work_base import ClientWorkBaseAttributes

class ServerAgentBaseDevicePlacementWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Server Agent Base Device Placement Work"
    device_object_location = Attribute('Device Object Location', min_version='21.4')
    device_share_mode = Attribute('Device Share Mode', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.metadata_base import MetadataBaseAttributes

class MetadataDateTimeAttributes(MetadataBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Metadata DateTime"
    date_only = Attribute('Date Only', min_version='21.4')



from __future__ import annotations

from typing import (
    Optional,
    Union,
)

from packaging.version import (
    parse as parse_version,
    Version,
)

from pyvenafi.tpp.api.session import Session
from pyvenafi.tpp.api.websdk.endpoints.authorize import _Authorize
from pyvenafi.tpp.api.websdk.endpoints.bus_status import _BusStatus
from pyvenafi.tpp.api.websdk.endpoints.certificates import _Certificates
from pyvenafi.tpp.api.websdk.endpoints.client import _Client
from pyvenafi.tpp.api.websdk.endpoints.codesign import _Codesign
from pyvenafi.tpp.api.websdk.endpoints.config import _Config
from pyvenafi.tpp.api.websdk.endpoints.config_schema import _ConfigSchema
from pyvenafi.tpp.api.websdk.endpoints.credentials import _Credentials
from pyvenafi.tpp.api.websdk.endpoints.crypto import _Crypto
from pyvenafi.tpp.api.websdk.endpoints.discovery import _Discovery
from pyvenafi.tpp.api.websdk.endpoints.flow import _Flow
from pyvenafi.tpp.api.websdk.endpoints.hsm_api import _HSMAPI
from pyvenafi.tpp.api.websdk.endpoints.identity import _Identity
from pyvenafi.tpp.api.websdk.endpoints.log import _Log
from pyvenafi.tpp.api.websdk.endpoints.metadata import _Metadata
from pyvenafi.tpp.api.websdk.endpoints.oauth import _Oauth
from pyvenafi.tpp.api.websdk.endpoints.permissions import _Permissions
from pyvenafi.tpp.api.websdk.endpoints.pki import _PKI
from pyvenafi.tpp.api.websdk.endpoints.platform import _Platform
from pyvenafi.tpp.api.websdk.endpoints.preferences import _Preferences
from pyvenafi.tpp.api.websdk.endpoints.processing_engines import _ProcessingEngines
from pyvenafi.tpp.api.websdk.endpoints.recycle_bin import _RecycleBin
from pyvenafi.tpp.api.websdk.endpoints.revoke import _Revoke
from pyvenafi.tpp.api.websdk.endpoints.secret_store import _SecretStore
from pyvenafi.tpp.api.websdk.endpoints.server_status import _ServerStatus
from pyvenafi.tpp.api.websdk.endpoints.ssh import _SSH
from pyvenafi.tpp.api.websdk.endpoints.ssh_certificates import _SSHCertificates
from pyvenafi.tpp.api.websdk.endpoints.stats import _Stats
from pyvenafi.tpp.api.websdk.endpoints.system_status import _SystemStatus
from pyvenafi.tpp.api.websdk.endpoints.teams import _Teams
from pyvenafi.tpp.api.websdk.endpoints.workflow import _Workflow
from pyvenafi.tpp.api.websdk.endpoints.x509_certificate_store import _X509CertificateStore
from pyvenafi.tpp.api.websdk.enums.oauth import Scope

_TPP_VERSION: Optional[Version] = None

class WebSDK:
    """
    Initializes a WebSDK session via username/password. Certificate authentication is not
    currently supported. Re-authentication occurs automatically when the API Key
    becomes invalidated. When initialized, all endpoints are also initialized.
    """

    def __init__(
        self, host: str, username: str, password: str, token: str = None, application_id: str = None,
        scope: Union[Scope, str] = None, refresh_token: str = None, proxies: dict = None,
        certificate_path: str = None, key_file_path: str = None, verify_ssl: bool = False,
        connection_timeout: float = None, read_timeout: float = None
    ):
        """
        Authenticates the given user to WebSDK. The only supported method for authentication at this time
        is with a username and password. Either an OAuth bearer token can be obtained, which requires
        both an Application ID and scope to be supplied, or the X-Venafi-API-Key can be obtained, which has
        been deprecated since TPP version 20.1.

        Using the OAuth authentication method requires an API Application Integration to be created that
        defines the maximum possible scope and the users/groups that are authorized to use that scope. That
        can be accomplished through Aperture. The ``scope`` parameter simply has to define the subset of
        allowed scopes defined by that application.

        Args:
            host: Hostname or IP Address of TPP
            username: Username
            password: Password
            token: OAuth Access Bearer Token
            application_id: Application ID of the OAuth API Application Integration. Must supply ``scope``.
            scope: Scope of the OAuth API Application Integration to be used. Must supply ``application_id``.
            proxies: An OrderedDict used by the python Requests library.
            certificate_path: Absolute path to the public certificate file.
            key_file_path: Absolute path to the private key file.
            verify_ssl: If ``True``, verify the SSL certificate of the target endpoints.
            connection_timeout: Timeout in seconds to establish a connection to the API service.
            read_timeout: Timeout in seconds between each byte received from the server.
        """
        # region Instance Variables
        self._host: str = host
        self._username: str = username
        self._password: str = password
        self._application_id: str = application_id
        self._scope: str = scope.to_string() if isinstance(scope, Scope) else scope
        self._token: str = token
        self._proxies: dict = proxies
        self._certificate_path: str = certificate_path
        self._key_file_path: str = key_file_path
        self._verify_ssl: bool = verify_ssl
        self._connection_timeout: float = connection_timeout
        self._read_timeout: float = read_timeout
        self._refresh_token: str = refresh_token

        # This is used by the endpoints to avoid redundancy.
        self._scheme = 'https'
        self._base_url = f'{self._scheme}://{host}'
        self._app_url = f'{self._base_url}/vedsdk'
        # endregion Instance Variables

        # region Authentication
        self._session = Session(
            headers={
                'Content-Type': 'application/json'
            },
            proxies=self._proxies,
            certificate_path=certificate_path,
            key_file_path=key_file_path,
            verify_ssl=verify_ssl,
            connection_timeout=connection_timeout,
            read_timeout=read_timeout
        )

        # Authorize the WebSDK session and store the API token.
        self.Authorize = _Authorize(self)
        self._authenticate()
        # endregion Authentication

        # region Initialize All WebSDK Endpoints
        # Initialize the rest of the endpoints with self, which contains the base url,
        # the authorization token, and the re-authentication method.
        self.BusStatus = _BusStatus(self)
        self.Certificates = _Certificates(self)
        self.Client = _Client(self)
        self.Codesign = _Codesign(self)
        self.Config = _Config(self)
        self.ConfigSchema = _ConfigSchema(self)
        self.Credentials = _Credentials(self)
        self.Crypto = _Crypto(self)
        self.Discovery = _Discovery(self)
        self.Flow = _Flow(self)
        self.HSMAPI = _HSMAPI(self)
        self.Identity = _Identity(self)
        self.Log = _Log(self)
        self.Metadata = _Metadata(self)
        self.Oauth = _Oauth(self)
        self.Permissions = _Permissions(self)
        self.PKI = _PKI(self)
        self.Platform = _Platform(self)
        self.Preferences = _Preferences(self)
        self.ProcessingEngines = _ProcessingEngines(self)
        self.RecycleBin = _RecycleBin(self)
        self.Revoke = _Revoke(self)
        self.SecretStore = _SecretStore(self)
        self.ServerStatus = _ServerStatus(self)
        self.SSH = _SSH(self)
        self.SSHCertificates = _SSHCertificates(self)
        self.Stats = _Stats(self)
        self.SystemStatus = _SystemStatus(self)
        self.Teams = _Teams(self)
        self.Workflow = _Workflow(self)
        self.X509CertificateStore = _X509CertificateStore(self)
        # endregion Initialize All WebSDK Endpoints

        # region Set TPP Version
        self._session.tpp_version = Version(f'{self.tpp_version.major}.{self.tpp_version.minor}')
        # endregion Set TPP Version

    @property
    def tpp_version(self):
        global _TPP_VERSION
        if _TPP_VERSION:
            return _TPP_VERSION
        _TPP_VERSION = parse_version(self.SystemStatus.Version.get().version)
        return _TPP_VERSION

    def _authenticate(self):
        if self._refresh_token:
            oauth = self.Authorize.Token.post(
                client_id=self._application_id,
                refresh_token=self._refresh_token
            )
            self._token = f'Bearer {oauth.access_token}'
            self._refresh_token = oauth.refresh_token
        elif not self._token:
            if self._application_id:
                if not self._scope:
                    raise ValueError(
                        f'OAuth authentication requires both an Application ID and a scope. '
                        f'The scope was not defined.'
                    )
                if self._certificate_path and self._key_file_path:
                    oauth = self.Authorize.Certificate.post(
                        client_id=self._application_id,
                        scope=self._scope
                    )
                else:
                    oauth = self.Authorize.OAuth.post(
                        username=self._username,
                        password=self._password,
                        client_id=self._application_id,
                        scope=self._scope
                    )
                self._token = f'Bearer {oauth.access_token}'
                self._refresh_token = oauth.refresh_token
            else:
                # Deprecated since TPP Version 20.1
                self._token = self.Authorize.post(username=self._username, password=self._password).token

        if self._token.endswith('=='):
            authorization_header = {
                'Authorization': self._token
            }
        else:
            authorization_header = {
                'X-Venafi-API-Key': self._token
            }
        self._session.update_headers(authorization_header)

    def re_authenticate(self):
        """
        Performs a re-authentication using the same parameters used to authorize initially.
        """
        self._token = None
        self._authenticate()



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class ReportFilterBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Report Filter Base"
    address = Attribute('Address', min_version='21.4')
    discoverydn = Attribute('DiscoveryDN', min_version='21.4')
    grouping = Attribute('Grouping', min_version='21.4')
    longrunning = Attribute('LongRunning', min_version='21.4')
    policydn = Attribute('PolicyDN', min_version='21.4')
    reporton = Attribute('ReportOn', min_version='21.4')
    selectedcontacts = Attribute('SelectedContacts', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class UpgradesRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Upgrades Root"
    grouping_id = Attribute('Grouping Id', min_version='21.4')
    origin_version = Attribute('Origin Version', min_version='21.4')
    start_time = Attribute('Start Time', min_version='21.4')
    stop_time = Attribute('Stop Time', min_version='21.4')
    target_version = Attribute('Target Version', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes

class ComodoCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Comodo CA"
    address = Attribute('Address', min_version='21.4')
    company_number = Attribute('Company Number', min_version='21.4')
    domain_control_validation = Attribute('Domain Control Validation', min_version='21.4')
    domain_control_validation_email = Attribute('Domain Control Validation Email', min_version='21.4')
    postal_code = Attribute('Postal Code', min_version='21.4')
    uri = Attribute('URI', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class ConnectionBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Connection Base"
    concurrent_connection_limit = Attribute('Concurrent Connection Limit', min_version='21.4')
    connection_method = Attribute('Connection Method', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    enforce_known_host = Attribute('Enforce Known Host', min_version='21.4')
    global_sudo = Attribute('Global sudo', min_version='21.4')
    host = Attribute('Host', min_version='21.4')
    last_known_fingerprint = Attribute('Last Known Fingerprint', min_version='21.4')
    last_known_key_type = Attribute('Last Known Key Type', min_version='21.4')
    port = Attribute('Port', min_version='21.4')
    remote_server_type = Attribute('Remote Server Type', min_version='21.4')
    secondary_credential = Attribute('Secondary Credential', min_version='21.4')
    sudo_password_delay = Attribute('Sudo Password Delay', min_version='21.4')
    temp_directory = Attribute('Temp Directory', min_version='21.4')
    terminal_columns = Attribute('Terminal Columns', min_version='21.4')
    terminal_rows = Attribute('Terminal Rows', min_version='21.4')
    terminal_type = Attribute('Terminal Type', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')
    trusted_fingerprint = Attribute('Trusted Fingerprint', min_version='21.4')
    trusted_key_type = Attribute('Trusted Key Type', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class LayoutRuleBaseAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Layout Rule Base"
    rule = Attribute('Rule')
    vault_id = Attribute('Vault Id')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.x509_certificate_base import X509CertificateBaseAttributes
from pyvenafi.tpp.attributes.x509_certificate_validation import X509CertificateValidationAttributes

class X509CertificateAttributes(
    X509CertificateBaseAttributes,
    X509CertificateValidationAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "X509 Certificate"
    acme_account_dn = Attribute('ACME Account DN', min_version='21.4')
    application_group_dn = Attribute('Application Group DN', min_version='21.4')
    custom_subjectaltname_othername_definition = Attribute(
        'Custom SubjectAltName OtherName Definition',
        min_version='24.1'
    )
    microsoft_ca_pool_certificate_authority = Attribute('Microsoft CA Pool:Certificate Authority', min_version='21.4')
    portal_download_count = Attribute('Portal Download Count', min_version='21.4')
    prohibited_san_types = Attribute('Prohibited SAN Types', min_version='21.4')
    sid_extension_allow_sid_outside_of_connected_identities = Attribute(
        'SID Extension:Allow SID Outside Of Connected Identities',
        min_version='23.1'
    )
    sid_extension_allowed = Attribute('SID Extension:Allowed', min_version='23.1')
    sid_extension_do_not_automatically_include_requester_identity = Attribute(
        'SID Extension:Do Not Automatically Include Requester Identity',
        min_version='23.1'
    )
    sid_extension_effective_value = Attribute('SID Extension:Effective Value', min_version='23.1')
    sid_extension_value = Attribute('SID Extension:Value', min_version='23.1')
    ticket_dn = Attribute('Ticket DN', min_version='21.4')
    work_dn = Attribute('Work DN', min_version='21.4')



from __future__ import annotations

from dataclasses import dataclass

from packaging.version import Version

@dataclass
class Permissions:
    name: str
    approve: bool = None
    delete: bool = None
    discover: bool = None
    manage: bool = None
    read: bool = None
    revoke: bool = None
    recyclebin: bool = None

    def to_dict(self):
        return {
            'approve'   : self.approve,
            'delete'    : self.delete,
            'discover'  : self.discover,
            'manage'    : self.manage,
            'read'      : self.read,
            'revoke'    : self.revoke,
            'recyclebin': self.recyclebin,
        }

    def effective(self):
        # "read" is implied when specifying the access permission, so it must be omitted.
        return [k for k, v in self.to_dict().items() if v and k != 'read']

@dataclass
class Scope:
    def __init__(self):
        self._certificate = Permissions(name='certificate')
        self._ssh = Permissions(name='ssh')
        self._codesign = Permissions(name='codesign')
        self._configuration = Permissions(name='configuration')
        self._restricted = Permissions(name='restricted')
        self._security = Permissions(name='security')
        self._statistics = Permissions(name='statistics')
        self._agent = Permissions(name='agent')
        self._admin = Permissions(name='admin')
        self._codesignclient = Permissions(name='codesignclient')

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    @property
    def scopes(self):
        return [
            {
                'name'       : p.name,
                'permissions': p.to_dict()
            }
            for p in self.list()
        ]

    def list(self):
        return [
            self._certificate, self._ssh, self._codesign,
            self._configuration, self._restricted,
            self._security, self._statistics, self._agent,
            self._admin, self._codesignclient
        ]

    def to_string(self):
        scopes = []
        for p in self.list():
            effective = p.effective()
            if len(effective) == 0:
                if p.read is True:
                    scopes.append(p.name)
                continue
            else:
                scopes.append(f'{p.name}:{",".join(effective)}')
        return ';'.join(scopes)

    def admin(self, delete: bool = False, read: bool = False, recyclebin: bool = False):
        self._admin.delete = delete
        self._admin.read = read
        self._admin.recyclebin = recyclebin
        return self

    def agent(self, delete: bool = False, read: bool = False):
        self._agent.delete = delete
        self._agent.read = read
        return self

    @classmethod
    def _all(
        cls,
        version: str = "22.1.0",
        admin: bool = True,
        agent: bool = True,
        certificate: bool = True,
        configuration: bool = True,
        codesign: bool = True,
        codesignclient: bool = True,
        restricted: bool = True,
        security: bool = True,
        ssh: bool = True,
        statistics: bool = True
    ):
        self = cls()
        if admin:
            self.admin(delete=True, read=True, recyclebin=False)
            if Version(version=version) >= Version(version="22.2.0"):
                self._admin.recyclebin = True
        if agent:
            self.agent(delete=True, read=True)
        if certificate:
            self.certificate(approve=True, delete=True, discover=True, manage=True, read=True, revoke=True)
        if configuration:
            self.configuration(delete=True, manage=True, read=True)
        if codesignclient:
            self.codesignclient(read=True)
        if codesign:
            self.codesign(delete=True, manage=True, read=True)
        if restricted:
            self.restricted(delete=True, manage=True, read=True)
        if security:
            self.security(delete=True, manage=True, read=True)
        if ssh:
            self.ssh(approve=True, delete=True, discover=True, manage=True, read=True)
        if statistics:
            self.statistics(read=True)
        return self

    def codesignclient(self, read: bool = False):
        self._codesignclient.read = read
        return self

    def certificate(
        self, approve: bool = False, delete: bool = False, discover: bool = False,
        manage: bool = False, read: bool = False, revoke: bool = False
    ):
        self._certificate.approve = approve
        self._certificate.delete = delete
        self._certificate.discover = discover
        self._certificate.manage = manage
        self._certificate.read = read
        self._certificate.revoke = revoke
        return self

    def configuration(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._configuration.delete = delete
        self._configuration.manage = manage
        self._configuration.read = read
        return self

    def codesign(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._codesign.delete = delete
        self._codesign.manage = manage
        self._codesign.read = read
        return self

    def restricted(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._restricted.delete = delete
        self._restricted.manage = manage
        self._restricted.read = read
        return self

    def security(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._security.delete = delete
        self._security.manage = manage
        self._security.read = read
        return self

    def ssh(
        self, approve: bool = False, delete: bool = False, discover: bool = False,
        manage: bool = False, read: bool = False
    ):
        self._ssh.approve = approve
        self._ssh.delete = delete
        self._ssh.discover = discover
        self._ssh.manage = manage
        self._ssh.read = read
        return self

    def statistics(self, read: bool = False):
        self._statistics.read = read
        return self

def main():
    scope = Scope(). \
        certificate(approve=True). \
        statistics(read=True)
    print(scope.to_string())

if __name__ == '__main__':
    main()



from __future__ import annotations

from typing import (
    Union,
)

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import pki

class _PKI(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/PKI')
        self.HashiCorp = self._HashiCorp(api_obj=api_obj, url=f'{self._url}/HashiCorp')

    class _HashiCorp(WebSdkEndpoint):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.CA = self._CA(api_obj=self._api_obj, url=f'{self._url}/CA')
            self.Role = self._Role(api_obj=self._api_obj, url=f'{self._url}/Role')

        class _CA(WebSdkEndpoint):
            def Guid(self, guid: str):
                return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

            def get(self):
                class Output(WebSdkOutputModel):
                    pkis: list[pki.PKI] = ApiField(default_factory=list, alias='pkis')

                return generate_output(output_cls=Output, response=self._get())

            def post(
                self, certificate: Union[dict, pki.Certificate], folder_dn: str, pki_path: str, roles: list[str],
                create_certificate_authority: bool = True, create_pki_role: bool = False, crl_address: str = None,
                installation: Union[dict, pki.Installation] = None, key_algorithm: str = None, key_bit_size: str = None,
                ocsp_address: str = None
            ):
                body = {
                    'Certificate'               : certificate,
                    'CreateCertificateAuthority': create_certificate_authority,
                    'CreatePKIRole'             : create_pki_role,
                    'CRLAddress'                : crl_address,
                    'FolderDn'                  : folder_dn,
                    'Installation'              : installation,
                    'KeyAlgorithm'              : key_algorithm,
                    'KeyBitSize'                : key_bit_size,
                    'OCSPAddress'               : ocsp_address,
                    'PkiPath'                   : pki_path,
                    'Roles'                     : roles
                }

                class Output(WebSdkOutputModel):
                    certificate_dn: str = ApiField(alias='CertificateDN')
                    certificate_guid: str = ApiField(alias='CertificateGuid')
                    guid: str = ApiField(alias='Guid')

                return generate_output(output_cls=Output, response=self._post(data=body))

            class _Guid(WebSdkEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.Renew = self._Renew(api_obj=self._api_obj, url=f'{self._url}/Renew')

                def delete(self):
                    class Output(WebSdkOutputModel):
                        certificate_dn: str = ApiField(alias='CertificateDN')
                        certificate_guid: str = ApiField(alias='CertificateGuid')
                        guid: str = ApiField(alias='Guid')

                    return generate_output(output_cls=Output, response=self._delete())

                def get(self):
                    class Output(WebSdkOutputModel):
                        certificate: pki.Certificate = ApiField(alias='Certificate')
                        create_certificate_authority: bool = ApiField(alias='CreateCertificateAuthority')
                        create_pki_role: bool = ApiField(alias='CreatePKIRole')
                        folder_dn: str = ApiField(alias='FolderDn')
                        installation: pki.Installation = ApiField(alias='Installation')
                        pki_path: str = ApiField(alias='PkiPath')
                        roles: list[str] = ApiField(alias='Roles', default_factory=list)

                    return generate_output(output_cls=Output, response=self._get())

                def post(self):
                    class Output(WebSdkOutputModel):
                        certificate_dn: str = ApiField(alias='CertificateDN')
                        certificate_guid: str = ApiField(alias='CertificateGuid')
                        guid: str = ApiField(alias='Guid')

                    return generate_output(output_cls=Output, response=self._post(data={}))

                def put(
                    self,
                    folder_dn: str,
                    pki_path: str,
                    roles: list[str],
                    certificate: Union[dict, pki.Certificate] = None,
                    create_certificate_authority: bool = True,
                    create_pki_role: bool = False,
                    crl_address: str = None,
                    installation: Union[dict, pki.Installation] = None,
                    key_algorithm: str = None,
                    key_bit_size: str = None,
                    ocsp_address: str = None
                ):
                    body = {
                        'Certificate'               : certificate,
                        'CreateCertificateAuthority': create_certificate_authority,
                        'CreatePKIRole'             : create_pki_role,
                        'CRLAddress'                : crl_address,
                        'FolderDn'                  : folder_dn,
                        'Installation'              : installation,
                        'KeyAlgorithm'              : key_algorithm,
                        'KeyBitSize'                : key_bit_size,
                        'OCSPAddress'               : ocsp_address,
                        'PkiPath'                   : pki_path,
                        'Roles'                     : roles
                    }

                    class Output(WebSdkOutputModel):
                        certificate_dn: str = ApiField(alias='CertificateDN')
                        certificate_guid: str = ApiField(alias='CertificateGuid')
                        guid: str = ApiField(alias='Guid')

                    return generate_output(output_cls=Output, response=self._put(data=body))

                class _Renew(WebSdkEndpoint):
                    def post(self):
                        class Output(WebSdkOutputModel):
                            certificate_dn: str = ApiField(alias='CertificateDN')
                            certificate_guid: str = ApiField(alias='CertificateGuid')
                            guid: str = ApiField(alias='Guid')

                        return generate_output(output_cls=Output, response=self._post(data={}))

        class _Role(WebSdkEndpoint):
            def post(
                self, folder_dn: str, role_name: str, city: str = None, country: str = None,
                enhanced_key_usage: list[str] = None, key_algorithm: str = None, key_bit_size: str = None,
                organization: str = None, organizational_units: list[str] = None, state: str = None,
                whitelisted_domains: list[str] = None
            ):
                body = {
                    'City'               : city,
                    'Country'            : country,
                    'EnhancedKeyUsage'   : enhanced_key_usage,
                    'FolderDn'           : folder_dn,
                    'KeyAlgorithm'       : key_algorithm,
                    'KeyBitSize'         : key_bit_size,
                    'Organization'       : organization,
                    'OrganizationalUnits': organizational_units,
                    'RoleName'           : role_name,
                    'State'              : state,
                    'WhitelistedDomains' : whitelisted_domains
                }

                class Output(WebSdkOutputModel):
                    guid: str = ApiField(alias='Guid')

                return generate_output(output_cls=Output, response=self._post(data=body))

            def Guid(self, guid: str):
                return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

            class _Guid(WebSdkEndpoint):
                def delete(self):
                    class Output(WebSdkOutputModel):
                        guid: str = ApiField(alias='Guid')

                    return generate_output(output_cls=Output, response=self._delete())

                def get(self):
                    class Output(WebSdkOutputModel):
                        city: str = ApiField(alias='City')
                        country: str = ApiField(alias='Country')
                        enhanced_key_usage: str = ApiField(alias='EnhancedKeyUsage')
                        folder_dn: str = ApiField(alias='FolderDn')
                        guid: str = ApiField(alias='Guid')
                        key_algorithm: str = ApiField(alias='KeyAlgorithm')
                        key_bit_size: str = ApiField(alias='KeyBitSize')
                        organization: str = ApiField(alias='Organization')
                        organizational_units: list[str] = ApiField(default_factory=list, alias='OrganizationalUnits')
                        role_name: str = ApiField(alias='RoleName')
                        state: str = ApiField(alias='State')
                        whitelisted_domains: list[str] = ApiField(default_factory=list, alias='WhitelistedDomains')

                    return generate_output(output_cls=Output, response=self._get())

                def put(
                    self, city: str = None, country: str = None,
                    enhanced_key_usage: list[str] = None, key_algorithm: str = None, key_bit_size: str = None,
                    organization: str = None, organizational_units: list[str] = None, state: str = None,
                    whitelisted_domains: list[str] = None
                ):
                    body = {
                        'City'               : city,
                        'Country'            : country,
                        'EnhancedKeyUsage'   : enhanced_key_usage,
                        'KeyAlgorithm'       : key_algorithm,
                        'KeyBitSize'         : key_bit_size,
                        'Organization'       : organization,
                        'OrganizationalUnits': organizational_units,
                        'State'              : state,
                        'WhitelistedDomains' : whitelisted_domains
                    }

                    class Output(WebSdkOutputModel):
                        guid: str = ApiField(alias='Guid')

                    return generate_output(output_cls=Output, response=self._put(data=body))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes

class EntrustSecurityManagerCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Entrust Security Manager CA"
    certificate_type = Attribute('Certificate Type', min_version='15.3')
    create_entrust_user = Attribute('Create Entrust User')
    epf_credential = Attribute('EPF Credential')
    epf_credential_dn = Attribute('EPF Credential DN')
    enrollment_server_for_web_folder = Attribute('Enrollment Server for Web Folder')
    ini_file = Attribute('INI File')
    role = Attribute('Role', min_version='15.3')
    searchbase = Attribute('Searchbase')
    user_class_name = Attribute('User Class Name', min_version='15.3')



from __future__ import annotations

from pyvenafi.tpp import vtypes as Types  # noqa
from pyvenafi.tpp.api.authenticate import Authenticate  # noqa
from pyvenafi.tpp.api.websdk import (
    enums,
    models,
)  # noqa
from pyvenafi.tpp.api.websdk.enums.oauth import Scope  # noqa
from pyvenafi.tpp.dn import DN
from pyvenafi.tpp.features.definitions.attribute_values import AttributeValues  # noqa
from pyvenafi.tpp.features.definitions.attributes import Attributes  # noqa
from pyvenafi.tpp.features.definitions.classes import Classes as ClassNames  # noqa
from pyvenafi.tpp.features.definitions.features import Features  # noqa



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class TopAttributes(metaclass=IterableMeta):
    __config_class__ = "Top"
    contact = Attribute('Contact', min_version='21.4')
    created_by = Attribute('Created By', min_version='21.4')
    description = Attribute('Description', min_version='21.4')
    disabled = Attribute('Disabled', min_version='21.4')
    escalation_contact = Attribute('Escalation Contact', min_version='21.4')
    guid = Attribute('GUID', min_version='21.4')
    managed_by = Attribute('Managed By', min_version='21.4')
    metadata = Attribute('Metadata', min_version='21.4')
    recycle_automatic = Attribute('Recycle Automatic', min_version='22.2')
    recycle_deleted_by = Attribute('Recycle Deleted By', min_version='22.2')
    recycle_deleted_on = Attribute('Recycle Deleted On', min_version='22.2')
    recycle_origin_cn = Attribute('Recycle Origin CN', min_version='22.2')
    recycle_origin_dn = Attribute('Recycle Origin DN', min_version='22.2')
    recycle_origin_parent_cn = Attribute('Recycle Origin Parent CN', min_version='22.2')
    recycle_origin_parent_id = Attribute('Recycle Origin Parent ID', min_version='22.2')
    recycle_origin_parent_pdn = Attribute('Recycle Origin Parent PDN', min_version='22.2')
    reference = Attribute('Reference', min_version='21.4')
    workflow = Attribute('Workflow', min_version='21.4')
    workflow_block = Attribute('Workflow Block', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Self(ObjectModel):
    href: str = ApiField(alias='Href')

class Link(ObjectModel):
    self: Self = ApiField(alias='Self')

class Engine(ObjectModel):
    links: Link = ApiField(alias='_links')
    engine_dn: str = ApiField(alias='EngineDn')
    engine_guid: str = ApiField(alias='EngineGuid')
    engine_name: str = ApiField(alias='EngineName')

class Folder(ObjectModel):
    folder_dn: str = ApiField(alias='FolderDN')
    folder_guid: str = ApiField(alias='FolderGuid')
    folder_name: str = ApiField(alias='FolderName')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.monitoring_base import MonitoringBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class KeyBaseAttributes(MonitoringBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Key Base"
    algorithm = Attribute('Algorithm')
    approver = Attribute('Approver')
    consumers = Attribute('Consumers')
    disable_automatic_renewal = Attribute('Disable Automatic Renewal')
    in_error = Attribute('In Error')
    internet_email_address = Attribute('Internet EMail Address')
    key_storage_location = Attribute('Key Storage Location')
    label = Attribute('Label')
    last_evaluated_on = Attribute('Last Evaluated On')
    last_notification = Attribute('Last Notification')
    last_renewed_by = Attribute('Last Renewed By')
    last_renewed_on = Attribute('Last Renewed On')
    management_type = Attribute('Management Type')
    max_uses = Attribute('Max Uses')
    notified_on = Attribute('Notified On')
    origin = Attribute('Origin')
    private_key_vault_id = Attribute('Private Key Vault Id')
    real_name = Attribute('Real Name')
    status = Attribute('Status')
    use_count = Attribute('Use Count')
    validity_period = Attribute('Validity Period')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class TreeRootAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Tree Root"
    company_name = Attribute('Company Name')
    migration_task = Attribute('Migration Task')
    pendo_eula_version = Attribute('Pendo EULA Version')
    pendo_optional_data_collection = Attribute('Pendo Optional Data Collection')
    schema_applied = Attribute('Schema Applied')
    schema_version = Attribute('Schema Version')
    usage_tracking = Attribute('Usage Tracking')
    use_company_name_for_analytics = Attribute('Use Company Name for Analytics')
    version = Attribute('Version')



from __future__ import annotations

from typing import (
    Union,
)

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.enums.config import CodeSignAttributeValues
from pyvenafi.tpp.api.websdk.models import codesign
from pyvenafi.tpp.api.websdk.models.codesign import CertificateEnvironment_Pre22_4

class _Codesign(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Codesign')
        self.AddAdministrator = self._AddAdministrator(api_obj=api_obj, url=f'{self._url}/AddAdministrator')
        self.AddApplicationAdministrator = self._AddApplicationAdministrator(
            api_obj=api_obj,
            url=f'{self._url}/AddApplicationAdministrator'
        )
        self.AddPreApproval = self._AddPreApproval(api_obj=api_obj, url=f'{self._url}/AddPreApproval')
        self.AddProjectApprover = self._AddProjectApprover(api_obj=api_obj, url=f'{self._url}/AddProjectApprover')
        self.CountReferences = self._CountReferences(api_obj=api_obj, url=f'{self._url}/CountReferences')
        self.CreateApplication = self._CreateApplication(api_obj=api_obj, url=f'{self._url}/CreateApplication')
        self.CreateApplicationCollection = self._CreateApplicationCollection(
            api_obj=api_obj,
            url=f'{self._url}/CreateApplicationCollection'
        )
        self.CreateEnvironment = self._CreateEnvironment(api_obj=api_obj, url=f'{self._url}/CreateEnvironment')
        self.CreateProject = self._CreateProject(api_obj=api_obj, url=f'{self._url}/CreateProject')
        self.CreateTemplate = self._CreateTemplate(api_obj=api_obj, url=f'{self._url}/CreateTemplate')
        self.DeleteApplication = self._DeleteApplication(api_obj=api_obj, url=f'{self._url}/DeleteApplication')
        self.DeleteApplicationCollection = self._DeleteApplicationCollection(
            api_obj=api_obj,
            url=f'{self._url}/DeleteApplicationCollection'
        )
        self.DeleteEnvironment = self._DeleteEnvironment(api_obj=api_obj, url=f'{self._url}/DeleteEnvironment')
        self.DeleteProject = self._DeleteProject(api_obj=api_obj, url=f'{self._url}/DeleteProject')
        self.DeleteTemplate = self._DeleteTemplate(api_obj=api_obj, url=f'{self._url}/DeleteTemplate')
        self.EnumerateApplicationCollections = self._EnumerateApplicationCollections(
            api_obj=api_obj,
            url=f'{self._url}/EnumerateApplicationCollections'
        )
        self.EnumerateApplications = self._EnumerateApplications(
            api_obj=api_obj,
            url=f'{self._url}/EnumerateApplications'
        )
        self.EnumerateProjects = self._EnumerateProjects(api_obj=api_obj, url=f'{self._url}/EnumerateProjects')
        self.EnumerateReferences = self._EnumerateReferences(api_obj=api_obj, url=f'{self._url}/EnumerateReferences')
        self.EnumerateTemplates = self._EnumerateTemplates(api_obj=api_obj, url=f'{self._url}/EnumerateTemplates')
        self.ExportSignArchive = self._ExportSignArchive(api_obj=api_obj, url=f'{self._url}/ExportSignArchive')
        self.FindEnvironment = self._FindEnvironment(api_obj=api_obj, url=f'{self._url}/FindEnvironment')
        self.GetApplication = self._GetApplication(api_obj=api_obj, url=f'{self._url}/GetApplication')
        self.GetApplicationCollection = self._GetApplicationCollection(
            api_obj=api_obj,
            url=f'{self._url}/GetApplicationCollection'
        )
        self.GetApplicationCollectionMemberDNs = self._GetApplicationCollectionMemberDNs(
            api_obj=api_obj,
            url=f'{self._url}/GetApplicationCollectionMemberDNs'
        )
        self.GetApplicationCollectionMembers = self._GetApplicationCollectionMembers(
            api_obj=api_obj,
            url=f'{self._url}/GetApplicationCollectionMembers'
        )
        self.GetEnvironment = self._GetEnvironment(api_obj=api_obj, url=f'{self._url}/GetEnvironment')
        self.GetGlobalConfiguration = self._GetGlobalConfiguration(
            api_obj=api_obj,
            url=f'{self._url}/GetGlobalConfiguration'
        )
        self.GetObjectRights = self._GetObjectRights(api_obj=api_obj, url=f'{self._url}/GetObjectRights')
        self.GetProject = self._GetProject(api_obj=api_obj, url=f'{self._url}/GetProject')
        self.GetRight = self._GetRight(api_obj=api_obj, url=f'{self._url}/GetRight')
        self.GetTemplate = self._GetTemplate(api_obj=api_obj, url=f'{self._url}/GetTemplate')
        self.GetTrusteeRights = self._GetTrusteeRights(api_obj=api_obj, url=f'{self._url}/GetTrusteeRights')
        self.RemoveAdministrator = self._RemoveAdministrator(api_obj=api_obj, url=f'{self._url}/RemoveAdministrator')
        self.RemoveApplicationAdministrator = self._RemoveApplicationAdministrator(
            api_obj=api_obj,
            url=f'{self._url}/RemoveApplicationAdministrator'
        )
        self.RemoveProjectApprover = self._RemoveProjectApprover(
            api_obj=api_obj,
            url=f'{self._url}/RemoveProjectApprover'
        )
        self.RenameApplication = self._RenameApplication(api_obj=api_obj, url=f'{self._url}/RenameApplication')
        self.RenameApplicationCollection = self._RenameApplicationCollection(
            api_obj=api_obj,
            url=f'{self._url}/RenameApplicationCollection'
        )
        self.RenameProject = self._RenameProject(api_obj=api_obj, url=f'{self._url}/RenameProject')
        self.RenameTemplate = self._RenameTemplate(api_obj=api_obj, url=f'{self._url}/RenameTemplate')
        self.RenewEnvironment = self._RenewEnvironment(api_obj=api_obj, url=f'{self._url}/RenewEnvironment')
        self.RetrieveArchiveEntries = self._RetrieveArchiveEntries(
            api_obj=api_obj,
            url=f'{self._url}/RetrieveArchiveEntries'
        )
        self.SetGlobalConfiguration = self._SetGlobalConfiguration(
            api_obj=api_obj,
            url=f'{self._url}/SetGlobalConfiguration'
        )
        self.UpdateApplication = self._UpdateApplication(api_obj=api_obj, url=f'{self._url}/UpdateApplication')
        self.UpdateApplicationCollection = self._UpdateApplicationCollection(
            api_obj=api_obj,
            url=f'{self._url}/UpdateApplicationCollection'
        )
        self.UpdateEnvironment = self._UpdateEnvironment(api_obj=api_obj, url=f'{self._url}/UpdateEnvironment')
        self.UpdateProject = self._UpdateProject(api_obj=api_obj, url=f'{self._url}/UpdateProject')
        self.UpdateProjectStatus = self._UpdateProjectStatus(api_obj=api_obj, url=f'{self._url}/UpdateProjectStatus')
        self.UpdateTemplate = self._UpdateTemplate(api_obj=api_obj, url=f'{self._url}/UpdateTemplate')

    class _AddAdministrator(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _AddApplicationAdministrator(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _AddProjectApprover(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _AddPreApproval(WebSdkEndpoint):
        def post(
            self, dn: str, comment: str, user: str, hours: int = None, ip_address: str = None,
            signing_executable: str = None, single_use: bool = None, not_before: str = None, ):
            body = {
                'Dn'               : dn,
                'Comment'          : comment,
                'Hours'            : hours,
                'IPAddress'        : ip_address,
                'SigningExecutable': signing_executable,
                'SingleUse'        : single_use,
                'User'             : user,
                'NotBefore'        : not_before
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CountReferences(WebSdkEndpoint):
        def post(
            self, application: Union[dict, codesign.Application] = None,
            application_collection: Union[dict, codesign.ApplicationCollection] = None
        ):
            body = {
                'Application'          : application,
                'ApplicationCollection': application_collection
            }

            class Output(WebSdkOutputModel):
                count: int = ApiField(alias='Count')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CreateApplication(WebSdkEndpoint):
        def post(self, dn: str):
            body = {
                'Dn': dn
            }

            class Output(WebSdkOutputModel):
                application: codesign.Application = ApiField(alias='Application')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CreateApplicationCollection(WebSdkEndpoint):
        def post(self, dn: str):
            body = {
                'Dn': dn
            }

            class Output(WebSdkOutputModel):
                application: codesign.ApplicationCollection = ApiField(alias='ApplicationCollection')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CreateEnvironment(WebSdkEndpoint):
        def post(
            self, project: dict[str, Union[str, int]],
            template_type: Union[CodeSignAttributeValues.TemplateTypeKey, str],
            template: list[dict[str, str]],
            dn: str = None, environment_name: str = None,
            template_dn: str = None,
        ):
            body = {
                'Dn'             : dn,
                'EnvironmentName': environment_name,
                'Project'        : project,
                template_type    : template,
                'TemplateDn'     : template_dn
            }

            class Output(WebSdkOutputModel):
                apple_environment: codesign.AppleEnvironment = ApiField(alias='AppleEnvironment')
                certificate_environment: codesign.CertificateEnvironment = ApiField(alias='CertificateEnvironment')
                csp_environment: codesign.CSPEnvironment = ApiField(alias='CSPEnvironment')
                dot_net_environment: codesign.DotNetEnvironment = ApiField(alias='DotNetEnvironment')
                gpg_environment: codesign.GPGEnvironment = ApiField(alias='GPGEnvironment')
                key_pair_environment: codesign.KeyPairEnvironment = ApiField(alias='KeyPairEnvironment')
                apple_template: codesign.AppleTemplate = ApiField(alias='AppleTemplate')
                certificate_template: codesign.CertificateTemplate = ApiField(alias='CertificateTemplate')
                csp_template: codesign.CSPTemplate = ApiField(alias='CSPTemplate')
                dot_net_template: codesign.DotNetTemplate = ApiField(alias='DotNetTemplate')
                gpg_template: codesign.GPGTemplate = ApiField(alias='GPGTemplate')
                key_pair_template: codesign.KeyPairTemplate = ApiField(alias='KeyPairTemplate')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CreateProject(WebSdkEndpoint):
        def post(self, dn: str):
            body = {
                'Dn': dn
            }

            class Output(WebSdkOutputModel):
                project: codesign.Project = ApiField(alias='Project')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _CreateTemplate(WebSdkEndpoint):
        def post(self, dn: str, template_type: codesign.TemplateType = None, per_user: bool = None):
            body = {
                'Dn'          : dn,
                'TemplateType': template_type,
                'PerUser'     : per_user
            }

            class Output(WebSdkOutputModel):
                certificate_template: codesign.CertificateTemplate = ApiField(alias='CertificateTemplate')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DeleteApplication(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DeleteApplicationCollection(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DeleteEnvironment(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DeleteProject(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _DeleteTemplate(WebSdkEndpoint):
        def post(self, dn: str = None, force: bool = None, guid: str = None, id: int = None):
            body = {
                'Dn'   : dn,
                'Force': force,
                'Guid' : guid,
                'Id'   : id
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateApplications(WebSdkEndpoint):
        # noinspection ALL
        def post(self, filter: str = None):
            body = {
                'Filter': filter
            }

            class Output(WebSdkOutputModel):
                applications: list[codesign.Application] = ApiField(alias='Applications', default_factory=list)
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateApplicationCollections(WebSdkEndpoint):
        # noinspection ALL
        def post(self, filter: str = None):
            body = {
                'Filter': filter
            }

            class Output(WebSdkOutputModel):
                application_collections: list[codesign.ApplicationCollection] = ApiField(
                    alias='ApplicationCollections',
                    default_factory=list
                )
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateProjects(WebSdkEndpoint):
        # noinspection ALL
        def post(self, filter: str = None, rights: int = None):
            body = {
                'Filter': filter,
                'Rights': rights
            }

            class Output(WebSdkOutputModel):
                projects: list[codesign.Project] = ApiField(alias='Projects', default_factory=list)
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateReferences(WebSdkEndpoint):
        def post(
            self, application: Union[dict, codesign.Application] = None,
            application_collection: Union[dict] = None,
            application_dn: str = None, application_guid: str = None,
            collection_dn: str = None, collection_guid: str = None
        ):
            body = {
                'Application'          : application,
                'ApplicationCollection': application_collection,
                'ApplicationDn'        : application_dn,
                'ApplicationGuid'      : application_guid,
                'CollectionDn'         : collection_dn,
                'CollectionGuid'       : collection_guid
            }

            class Output(WebSdkOutputModel):
                reference_dns: list[str] = ApiField(alias='ReferenceDNs', default_factory=list)
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _EnumerateTemplates(WebSdkEndpoint):
        # noinspection ALL
        def post(self, filter: str = None):
            body = {
                'Filter': filter
            }

            class Output(WebSdkOutputModel):
                apple_templates: list[codesign.AppleTemplate] = ApiField(alias='AppleTemplates', default_factory=list)
                certificate_templates: list[codesign.CertificateTemplate] = ApiField(
                    alias='CertificateTemplates',
                    default_factory=list
                )
                csp_templates: list[codesign.CSPTemplate] = ApiField(alias='CSPTemplates', default_factory=list)
                dot_net_templates: list[codesign.DotNetTemplate] = ApiField(
                    alias='DotNetTemplates',
                    default_factory=list
                )
                gpg_templates: list[codesign.GPGTemplate] = ApiField(alias='GPGTemplates', default_factory=list)
                key_pair_templates: list[codesign.KeyPairTemplate] = ApiField(
                    alias='KeyPairTemplates',
                    default_factory=list
                )
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _ExportSignArchive(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(output_cls=Output, response=self._get())

    class _FindEnvironment(WebSdkEndpoint):
        def post(self, public_key_hash: str = None):
            body = {
                "PublicKeyHash": public_key_hash,
            }

            class Output(WebSdkOutputModel):
                project: codesign.Project = ApiField(alias='Project')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _GetApplication(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                application: codesign.Application = ApiField(alias='Application')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetApplicationCollection(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                application_collection: codesign.ApplicationCollection = ApiField(alias='ApplicationCollection')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetApplicationCollectionMembers(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                application_collection: codesign.ApplicationCollection = ApiField(alias='ApplicationCollection')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetApplicationCollectionMemberDNs(WebSdkEndpoint):
        def post(
            self,
            dn: str = None,
            guid: str = None,
            id: int = None,
            application: Union[dict, codesign.Application] = None
        ):
            body = {
                'Application': application,
                'Dn'         : dn,
                'Guid'       : guid,
                'Id'         : id
            }

            class Output(WebSdkOutputModel):
                application_collection: codesign.ApplicationCollection = ApiField(alias='ApplicationCollection')
                application_collection_dns: list[str] = ApiField(alias='ApplicationCollectionDNs', default_factory=list)
                application_dns: list[str] = ApiField(alias='ApplicationDNs', default_factory=list)
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetEnvironment(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                apple_environment: codesign.AppleEnvironment = ApiField(alias='AppleEnvironment')
                certificate_environment: codesign.CertificateEnvironment = ApiField(alias='CertificateEnvironment')
                csp_environment: codesign.CSPEnvironment = ApiField(alias='CSPEnvironment')
                dot_net_environment: codesign.DotNetEnvironment = ApiField(alias='DotNetEnvironment')
                gpg_environment: codesign.GPGEnvironment = ApiField(alias='GPGEnvironment')
                key_pair_environment: codesign.KeyPairEnvironment = ApiField(alias='KeyPairEnvironment')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetGlobalConfiguration(WebSdkEndpoint):
        def get(self):
            class Output(WebSdkOutputModel):
                global_configuration: codesign.GlobalConfiguration = ApiField(alias='GlobalConfiguration')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._get(), output_cls=Output)

    class _GetObjectRights(WebSdkEndpoint):
        def post(self, dn: str):
            body = {
                'Dn': dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                rights_list: list[codesign.RightsKeyValue] = ApiField(alias='RightsList', default_factory=list)
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetProject(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                project: codesign.Project = ApiField(alias='Project')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetRight(WebSdkEndpoint):
        def post(self, dn: str):
            body = {
                'Dn': dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                rights: codesign.Rights = ApiField(alias='Rights', converter=lambda x: codesign.Rights(value=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetTemplate(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'  : dn,
                'Guid': guid,
                'Id'  : id
            }

            class Output(WebSdkOutputModel):
                apple_template: codesign.AppleTemplate = ApiField(alias='AppleTemplate')
                certificate_template: codesign.CertificateTemplate = ApiField(alias='CertificateTemplate')
                csp_template: codesign.CSPTemplate = ApiField(alias='CSPTemplate')
                dot_net_template: codesign.DotNetTemplate = ApiField(alias='DotNetTemplate')
                gpg_template: codesign.GPGTemplate = ApiField(alias='GPGTemplate')
                key_pair_template: codesign.KeyPairTemplate = ApiField(alias='KeyPairTemplate')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _GetTrusteeRights(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                rights_list: list[codesign.RightsKeyValue] = ApiField(alias='RightsList', default_factory=list)
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemoveAdministrator(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemoveApplicationAdministrator(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RemoveProjectApprover(WebSdkEndpoint):
        def post(self, trustee: str):
            body = {
                'Trustee': trustee
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RenameApplication(WebSdkEndpoint):
        def post(self, dn: str, new_dn: str):
            body = {
                'Dn'   : dn,
                'NewDn': new_dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RenameApplicationCollection(WebSdkEndpoint):
        def post(self, dn: str, new_dn: str):
            body = {
                'Dn'   : dn,
                'NewDn': new_dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RenameProject(WebSdkEndpoint):
        def post(self, new_dn: str, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'   : dn,
                'Guid' : guid,
                'Id'   : id,
                'NewDn': new_dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RenameTemplate(WebSdkEndpoint):
        def post(self, new_dn: str, dn: str = None, guid: str = None):
            body = {
                'Dn'   : dn,
                'Guid' : guid,
                'Id'   : id,
                'NewDn': new_dn
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _RenewEnvironment(WebSdkEndpoint):
        def post(self, dn: str = None, guid: str = None, id: int = None):
            body = {
                "Dn"  : dn,
                "Guid": guid,
                "Id"  : id,
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _RetrieveArchiveEntries(WebSdkEndpoint):
        def post(
            self,
            archive_filter: Union[dict, codesign.ArchiveFilter],
            page_size: int = None,
            page: int = None
        ):
            body = {
                'ArchiveFilter': archive_filter,
                'PageSize'     : page_size,
                'Page'         : page
            }

            class Output(WebSdkOutputModel):
                archive_results: codesign.ArchiveResults = ApiField(alias='ArchiveResults')
                page_number: int = ApiField(alias='PageNumber')
                success: bool = ApiField(alias='Success')
                total_count: int = ApiField(alias='TotalCount')

            if not self._is_version_compatible(minimum='22.4'):
                self._log_warning_message(
                    'Cannot call endpoint POST Codesign/RetrieveArchiveEntries because the endpoint does '
                    'not exist before version 22.4.'
                )
            else:
                return generate_output(output_cls=Output, response=self._post(data=body))

    class _SetGlobalConfiguration(WebSdkEndpoint):
        def post(self, global_configuration: Union[dict, codesign.GlobalConfiguration]):
            body = {
                'GlobalConfiguration': global_configuration
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateApplication(WebSdkEndpoint):
        def post(self, application: Union[dict, codesign.Application]):
            body = {
                'Application': application
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateApplicationCollection(WebSdkEndpoint):
        def post(self, application_collection: Union[dict, codesign.ApplicationCollection]):
            body = {
                'ApplicationCollection': application_collection
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateEnvironment(WebSdkEndpoint):
        def post(
            self, certificate_environment: Union[dict, codesign.CertificateEnvironment] = None,
            apple_environment: Union[dict, codesign.AppleEnvironment] = None,
            csp_environment: Union[dict, codesign.CSPEnvironment] = None,
            dot_net_environment: Union[dict, codesign.DotNetEnvironment] = None,
            gpg_environment: Union[dict, codesign.GPGEnvironment] = None,
            key_pair_environment: Union[dict, codesign.KeyPairEnvironment] = None
        ):
            if isinstance(certificate_environment, codesign.CertificateEnvironment) \
                and not self._is_version_compatible(minimum='22.4'):
                certificate_environment = codesign.CertificateEnvironment_Pre22_4(**certificate_environment.dict())
            body = {
                'AppleEnvironment'      : apple_environment,
                'CertificateEnvironment': certificate_environment,
                'CSPEnvironment'        : csp_environment,
                'DotNetEnvironment'     : dot_net_environment,
                'GPGEnvironment'        : gpg_environment,
                'KeyPairEnvironment'    : key_pair_environment
            }

            class Output(WebSdkOutputModel):
                apple_environment: codesign.AppleEnvironment = ApiField(alias='AppleEnvironment')
                certificate_environment: codesign.CertificateEnvironment = ApiField(alias='CertificateEnvironment')
                csp_environment: codesign.CSPEnvironment = ApiField(alias='CSPEnvironment')
                dot_net_environment: codesign.DotNetEnvironment = ApiField(alias='DotNetEnvironment')
                gpg_environment: codesign.GPGEnvironment = ApiField(alias='GPGEnvironment')
                key_pair_environment: codesign.KeyPairEnvironment = ApiField(alias='KeyPairEnvironment')
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateProject(WebSdkEndpoint):
        def post(self, project: Union[dict, codesign.Project]):
            if isinstance(project, codesign.Project) and project.certificate_environments \
                and not self._is_version_compatible(minimum='22.4'):
                project.certificate_environments = [
                    CertificateEnvironment_Pre22_4(**c.dict()) for c in project.certificate_environments
                ]
            body = {
                'Project': project
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateProjectStatus(WebSdkEndpoint):
        def post(self, project_status: int, dn: str = None, guid: str = None, id: int = None):
            body = {
                'Dn'           : dn,
                'Guid'         : guid,
                'Id'           : id,
                'ProjectStatus': project_status,
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _UpdateTemplate(WebSdkEndpoint):
        def post(
            self, dn: str, certificate_template: Union[dict, codesign.CertificateTemplate, codesign.CSPTemplate,
            codesign.DotNetTemplate, codesign.AppleTemplate, codesign.GPGTemplate, codesign.KeyPairTemplate],
            object_naming_pattern: str = None
        ):
            body = {
                'Dn'                 : dn,
                'CertificateTemplate': certificate_template,
                'ObjectNamingPattern': object_naming_pattern
            }

            class Output(WebSdkOutputModel):
                result: codesign.ResultCode = ApiField(alias='Result', converter=lambda x: codesign.ResultCode(code=x))
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes

class GlobalSignMSSLCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
    __config_class__ = "GlobalSign MSSL CA"
    domain_id = Attribute('Domain ID', min_version='21.4')
    profile_id = Attribute('Profile ID', min_version='21.4')
    san_type = Attribute('SAN Type', min_version='21.4')
    web_service_url = Attribute('Web Service URL', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class JSSClusterAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "JSS Cluster"
    contact = Attribute('Contact')
    description = Attribute('Description')
    disabled = Attribute('Disabled')
    discovered_by_dn = Attribute('Discovered By DN')
    jss_cluster_name = Attribute('JSS Cluster Name')
    jss_cluster_type = Attribute('JSS Cluster Type')
    jss_last_seen = Attribute('JSS Last Seen')
    jss_status = Attribute('JSS Status')
    last_sync = Attribute('Last Sync')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ACMEAuthorizationAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "ACME Authorization"
    acme_auth_id_type = Attribute('ACME Auth ID Type', min_version='21.4')
    acme_auth_id_value = Attribute('ACME Auth ID Value', min_version='21.4')
    acme_challenge_dn = Attribute('ACME Challenge DN', min_version='21.4')
    acme_challenge_token = Attribute('ACME Challenge Token', min_version='21.4')
    acme_challenge_validated_on = Attribute('ACME Challenge Validated On', min_version='21.4')
    acme_expires = Attribute('ACME Expires', min_version='21.4')
    status = Attribute('Status', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.service_module import ServiceModuleAttributes

class CertificateManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
    __config_class__ = "Certificate Manager"
    cdp_aia_verification_disabled = Attribute('CDP AIA Verification Disabled', min_version='21.4')
    certificate_api_todo_timeout = Attribute('Certificate API ToDo Timeout', min_version='21.4')
    certificate_retrieve_cache = Attribute('Certificate Retrieve Cache', min_version='22.4')
    escalation_notice_interval = Attribute('Escalation Notice Interval', min_version='21.4')
    escalation_notice_start = Attribute('Escalation Notice Start', min_version='21.4')
    expiration_notice_interval = Attribute('Expiration Notice Interval', min_version='21.4')
    expiration_notice_start = Attribute('Expiration Notice Start', min_version='21.4')
    maximum_threads = Attribute('Maximum Threads', min_version='21.4')
    minimum_threads = Attribute('Minimum Threads', min_version='21.4')
    renewal_window = Attribute('Renewal Window', min_version='21.4')
    renewal_window_event = Attribute('Renewal Window Event', min_version='21.4')
    revocation_check_disabled = Attribute('Revocation Check Disabled', min_version='21.4')
    service_module_classes = Attribute('Service Module Classes', min_version='21.4')
    start_time = Attribute('Start Time', min_version='21.4')
    trust_store_management_disabled = Attribute('Trust Store Management Disabled', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class LayoutRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Layout Root"
    teams_template_code_signing_vault_id = Attribute('Teams Template Code Signing Vault Id', min_version='21.4')
    teams_template_ssh_vault_id = Attribute('Teams Template SSH Vault Id', min_version='21.4')
    teams_template_tls_vault_id = Attribute('Teams Template TLS Vault Id', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes
from pyvenafi.tpp.attributes.report_filter_base import ReportFilterBaseAttributes

class ReportTrustAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:Trust"
    discoverydn = Attribute('DiscoveryDN', min_version='21.4')
    grouping = Attribute('Grouping', min_version='21.4')
    options = Attribute('Options', min_version='21.4')
    trustedca = Attribute('TrustedCA', min_version='21.4')
    untrustedca = Attribute('UntrustedCA', min_version='21.4')



import logging
import time
import os
from typing import TYPE_CHECKING
from pyvenafi.logger import (
    logger,
    features_logger,
)

if TYPE_CHECKING:
    from pyvenafi.cloud.api.authenticate import Authenticate

def feature(name: str):
    def decorate(cls):
        if int(os.getenv('PYVAAS_DOC_IN_PROGRESS', 0)):
            return cls
        setattr(cls, '__feature__', name)
        return features_logger.wrap_class(
            level=logging.DEBUG,
            exclude='_.*'
        )(cls)

    return decorate

class FeatureBase:
    def __init__(self, api: 'Authenticate'):
        self._api = api

    @staticmethod
    def _log_warning_message(msg: str):
        features_logger.warning(msg, stacklevel=2)

    @staticmethod
    def __no_op(*args, **kwargs):
        pass

    class _Timeout:
        def __init__(self, timeout):
            self.timeout = timeout
            self.max_time = timeout + time.time()
            self._cm = logger.suppressed(999)

        def __enter__(self):
            self._cm.__enter__()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._cm.__exit__(exc_type, exc_val, exc_tb)
            return

        def is_expired(self, poll: float = 0.5):
            if time.time() >= self.max_time:
                return True
            if poll:
                time.sleep(poll)

        @staticmethod
        def poll(seconds: float):
            time.sleep(seconds)



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Identity(ObjectModel):
    full_name: str = ApiField(alias='FullName')
    is_container: bool = ApiField(alias='IsContainer')
    is_group: bool = ApiField(alias='IsGroup')
    name: str = ApiField(alias='Name')
    prefix: str = ApiField(alias='Prefix')
    prefixed_name: str = ApiField(alias='PrefixedName')
    prefixed_universal: str = ApiField(alias='PrefixedUniversal')
    state: int = ApiField(alias='State')
    type: int = ApiField(alias='Type')
    universal: str = ApiField(alias='Universal')

    @property
    def is_user(self):
        return self.type & 1 == 1

    @property
    def is_security_group(self):
        return self.type & 2 == 2

    @property
    def is_distribution_group(self):
        return self.type & 8 == 8

class InvalidIdentity(ObjectModel):
    prefix: str = ApiField(alias='Prefix')
    prefixed_name: str = ApiField(alias='PrefixedName')
    prefixed_universal: str = ApiField(alias='PrefixedUniversal')
    universal: str = ApiField(alias='Universal')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.client_work_base import ClientWorkBaseAttributes

class ClientCertificateDiscoveryWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Client Certificate Discovery Work"
    certificate_location_dn = Attribute('Certificate Location DN', min_version='21.4')
    certificate_scanner_capi = Attribute('Certificate Scanner CAPI', min_version='21.4')
    certificate_scanner_map = Attribute('Certificate Scanner Map', min_version='21.4')
    certificate_scanner_native = Attribute('Certificate Scanner Native', min_version='21.4')
    certificate_scanner_native_stores = Attribute('Certificate Scanner Native Stores', min_version='21.4')
    certificate_scanner_path = Attribute('Certificate Scanner Path', min_version='21.4')
    clear_cache_timestamp = Attribute('Clear Cache Timestamp', min_version='21.4')
    credential = Attribute('Credential', min_version='21.4')
    days_of_month = Attribute('Days Of Month', min_version='21.4')
    days_of_week = Attribute('Days Of Week', min_version='21.4')
    device_location_dn = Attribute('Device Location DN', min_version='21.4')
    discovery_container = Attribute('Discovery Container', min_version='21.4')
    exclude_remote_mount_points = Attribute('Exclude Remote Mount Points', min_version='21.4')
    interval = Attribute('Interval', min_version='21.4')
    log_threshold = Attribute('Log Threshold', min_version='21.4')
    max_filesize = Attribute('Max Filesize', min_version='21.4')
    minimize_resource_use = Attribute('Minimize Resource Use', min_version='21.4')
    naming_pattern = Attribute('Naming Pattern', min_version='21.4')
    placement_rule = Attribute('Placement Rule', min_version='21.4')
    placement_summary = Attribute('Placement Summary', min_version='21.4')
    schedule_type = Attribute('Schedule Type', min_version='21.4')
    start_time = Attribute('Start Time', min_version='21.4')
    treat_unknown_roots_as_self_signed = Attribute('Treat Unknown Roots As Self Signed', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Certificate(ObjectModel):
    certificate: str = ApiField(alias='certificate')
    fingerprint: str = ApiField(alias='fingerprint')

class Protocol(ObjectModel):
    certificates: list[str] = ApiField(alias='certificates')
    protocol: str = ApiField(alias='protocol')

class Endpoint(ObjectModel):
    certificates: list[Certificate] = ApiField(alias='certificates')
    host: str = ApiField(alias='host')
    ip: str = ApiField(alias='ip')
    port: int = ApiField(alias='port')
    protocols: list[Protocol] = ApiField(alias='protocols')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.organization import OrganizationAttributes

class PlacementJobContainerAttributes(OrganizationAttributes, metaclass=IterableMeta):
    __config_class__ = "Placement Job Container"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.adaptable_flow_action_base import AdaptableFlowActionBaseAttributes

class SSHCertificateAdaptableFlowActionAttributes(AdaptableFlowActionBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH Certificate Adaptable Flow Action"
    allow_to_modify_certificate_request = Attribute('Allow To Modify Certificate Request', min_version='21.4')
    certificate_credential = Attribute('Certificate Credential', min_version='21.4')
    ssh_certificate_adaptable_flow_action_text_field_1 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 1',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_10 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 10',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_11 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 11',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_12 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 12',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_2 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 2',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_3 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 3',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_4 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 4',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_5 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 5',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_6 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 6',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_7 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 7',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_8 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 8',
        min_version='21.4'
    )
    ssh_certificate_adaptable_flow_action_text_field_9 = Attribute(
        'SSH Certificate Adaptable Flow Action Text Field 9',
        min_version='21.4'
    )
    secondary_credential = Attribute('Secondary Credential', min_version='21.4')
    service_address = Attribute('Service Address', min_version='21.4')
    username_credential = Attribute('Username Credential', min_version='21.4')
    validate_certificate_request_before_issuance = Attribute(
        'Validate Certificate Request Before Issuance',
        min_version='21.4'
    )



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class VamSunimpAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "VamSunimp"
    file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
    key_list_path = Attribute('Key List Path')
    key_store_credential = Attribute('Key Store Credential')
    key_store_name = Attribute('Key Store Name')
    key_store_validation_disabled = Attribute('Key Store Validation Disabled')
    sca6000_utility_path = Attribute('SCA6000 Utility Path')
    sunimp_utility_path = Attribute('Sunimp Utility Path')
